'''
Created on 2016/1/5

:author: hubo
'''

from vlcp.protocol import Protocol
from vlcp.config import defaultconfig
from vlcp.event import Event, withIndices, ConnectionWriteEvent
import logging
import itertools
from vlcp.event.lock import Lock
from contextlib import closing
from vlcp.event.event import M_
try:
    import hiredis
    hiredis_available = True
except Exception:
    hiredis_available = False


@withIndices('state', 'connection', 'connmark', 'createby')
class RedisConnectionStateEvent(Event):
    CONNECTION_UP = 'up'
    CONNECTION_DOWN = 'down'
    CONNECTION_NOTCONNECTED = 'notconnected'


@withIndices('connection', 'connmark', 'id', 'iserror', 'createby')
class RedisResponseEvent(Event):
    pass


@withIndices('type', 'subscribe', 'connection', 'connmark', 'createby')
class RedisSubscribeEvent(Event):
    SUBSCRIBE = 'subscribe'
    UNSUBSCRIBE = 'unsubscribe'
    PSUBSCRIBE = 'psubscribe'
    PUNSUBSCRIBE = 'punsubscribe'


@withIndices('type', 'subscribe', 'channel', 'connection', 'connmark', 'createby')
class RedisSubscribeMessageEvent(Event):
    MESSAGE = 'message'
    PMESSAGE = 'pmessage'


class RedisProtocolException(Exception):
    pass


class RedisReplyException(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)
        if args:
            self.subtype, _, self.describe = args[0].partition(' ')
        else:
            self.subtype = ''
            self.describe = ''


def _copy(buffer):
    try:
        if isinstance(buffer, memoryview):
            return buffer.tobytes()
        else:
            return buffer[:]
    except Exception:
        return buffer[:]


def _str(b, encoding = 'ascii'):
    if isinstance(b, str):
        return b
    elif isinstance(b, bytes):
        return b.decode(encoding)
    else:
        return str(b)


class RedisParser(object):
    "Python implemented hiredis.Reader()"
    def __init__(self):
        self._buffer = b''
        self._parser = self._parser_gen()

    def feed(self, data):
        self._buffer += data

    def gets(self):
        try:
            return next(self._parser)
        except StopIteration:
            raise RedisProtocolException(*self._exceptionargs)
        except RedisProtocolException as exc:
            self._exceptionargs = exc.args
            raise

    def _parser_gen(self, startpos = 0):
        try:
            _blk_buffer = []
            while True:
                lastfind = startpos
                while True:
                    le = self._buffer.find(b'\r', lastfind, -1)
                    if le < 0:
                        lastfind = len(self._buffer) - 1
                        if lastfind < startpos:
                            lastfind = startpos
                        if startpos > 0:
                            self._buffer = self._buffer[startpos:]
                            lastfind -= startpos
                            startpos = 0
                        yield False
                    else:
                        break
                c = self._buffer[startpos : startpos + 1]
                if c == b'+':
                    r = startpos + 1
                    startpos = le + 2
                    yield self._buffer[r : le]
                elif c == b'-':
                    r = startpos + 1
                    startpos = le + 2
                    yield RedisReplyException(_str(self._buffer[r : le]))
                elif c == b':':
                    r = startpos + 1
                    startpos = le + 2
                    try:
                        yield int(self._buffer[r : le])
                    except ValueError as exc:
                        raise RedisProtocolException(str(exc))
                elif c == b'$':
                    try:
                        blksize = int(self._buffer[startpos + 1 : le])
                    except ValueError as exc:
                        raise RedisProtocolException(str(exc))
                    startpos = le + 2
                    if blksize < 0:
                        yield None
                    else:
                        while len(self._buffer) - startpos < blksize:
                            _blk_buffer.append(self._buffer[startpos:])
                            blksize -= len(self._buffer) - startpos
                            self._buffer = b''
                            startpos = 0
                            yield False
                        _blk_buffer.append(self._buffer[startpos:blksize+startpos])
                        startpos += blksize
                        while len(self._buffer) - startpos < 2:
                            yield False
                        startpos += 2
                        r = b''.join(_blk_buffer)
                        del _blk_buffer[:]
                        yield r
                elif c == b'*':
                    try:
                        arraysize = int(self._buffer[startpos + 1 : le])
                    except ValueError as exc:
                        raise RedisProtocolException(str(exc))
                    startpos = le + 2
                    if arraysize < 0:
                        yield None
                    else:
                        array = []
                        if arraysize:
                            p = self._parser_gen(startpos)
                            try:
                                while arraysize:
                                    nv = next(p)
                                    if nv is not False:
                                        array.append(nv)
                                        arraysize -= 1
                                    else:
                                        yield False
                            finally:
                                p.close()
                                startpos = self._lastpos
                        yield array
                else:
                    raise RedisProtocolException(repr(c) + ' is not a valid RESP type')
        finally:
            self._lastpos = startpos


@defaultconfig
class Redis(Protocol):
    '''
    Redis (RESP) Protocol
    '''
    _default_persist = True
    # Default Redis server port
    _default_defaultport = 6379
    _default_createqueue = True
    # Limit response messages queue size
    _default_messagequeuesize = 4096
    # Default limit Redis bulk string to 64MB
    _default_bulklimit = 67108864
    # Default limit Redis array level to 128 levels
    _default_levellimit = 128
    # Default encoding when using Unicode strings in Redis commands
    _default_encoding = 'utf-8'
    # Use hiredis if possible
    _default_hiredis = True
    # Send PING command when the connection is idle
    _default_keepalivetime = 10
    # Disconnect when PING command does not have response
    _default_keepalivetimeout = 5
    # Connect timeout
    _default_connect_timeout = 5
    _default_tcp_nodelay = True
    _logger = logging.getLogger(__name__ + '.Redis')

    def __init__(self):
        '''
        Constructor
        '''
        Protocol.__init__(self)
        self.usehiredis = hiredis_available and self.hiredis
        self._format_request_init(self.encoding)

    async def init(self, connection):
        await Protocol.init(self, connection)
        connection.createdqueues.append(connection.scheduler.queue.addSubQueue(\
                self.messagepriority - 1, RedisConnectionStateEvent.createMatcher(connection = connection), ('connstate', connection)))
        connection.createdqueues.append(connection.scheduler.queue.addSubQueue(\
                self.messagepriority + 1, RedisResponseEvent.createMatcher(connection = connection), ('response', connection), self.messagequeuesize))
        connection.createdqueues.append(connection.scheduler.queue.addSubQueue(\
                self.messagepriority + 1, RedisSubscribeEvent.createMatcher(connection = connection), ('subscribe', connection), self.messagequeuesize))
        connection.createdqueues.append(connection.scheduler.queue.addSubQueue(\
                self.messagepriority, RedisSubscribeMessageEvent.createMatcher(connection = connection), ('message', connection), self.messagequeuesize))
        connection.redis_subscribe = False
        connection.redis_select = None
        connection.redis_subscribe_keys = set()
        connection.redis_subscribe_pkeys = set()
        await self.reconnect_init(connection)

    def _format_request_init(self, encoding):
        def _bytes(a):
            if isinstance(a, bytes):
                return a
            else:
                return str(a).encode(encoding)
        def format_request(*args):
            return b'*' + str(len(args)).encode('ascii') + b'\r\n' + b''.join(itertools.chain.from_iterable((b'$',str(len(sa)).encode('ascii'),b'\r\n',sa,b'\r\n') for sa in (_bytes(a) for a in args)))
        self._bytes = _bytes
        self.format_request = format_request

    async def reconnect_init(self, connection):
        connection.xid = 1
        connection.redis_replyxid = 1
        connection.redis_ping = -1
        connection.redis_pingreply = -1
        connection.redis_bufferedxid = 0
        connection.redis_sendbuffer = []
        connection.redis_sender = False
        write_buffer = []
        if self.usehiredis:
            connection.redis_reader = hiredis.Reader(protocolError = RedisProtocolException, replyError = RedisReplyException)
        else:
            connection.redis_reader = RedisParser()
        if connection.redis_select:
            write_buffer.append(self.format_request(b'SELECT', connection.redis_select))
            connection.xid += 1
        if connection.redis_subscribe:
            if connection.redis_subscribe_keys:
                write_buffer.append(self.format_request(b'SUBSCRIBE', *tuple(connection.redis_subscribe_keys)))
            if connection.redis_subscribe_pkeys:
                write_buffer.append(self.format_request(b'PSUBSCRIBE', *tuple(connection.redis_subscribe_pkeys)))
        connection.scheduler.emergesend(ConnectionWriteEvent(connection, connection.connmark, data=b''.join(write_buffer)))
        await connection.wait_for_send(RedisConnectionStateEvent(RedisConnectionStateEvent.CONNECTION_UP, connection, connection.connmark, self))

    async def closed(self, connection):
        await Protocol.closed(self, connection)
        self._logger.info('Redis connection is closed on %r', connection)
        await connection.wait_for_send(RedisConnectionStateEvent(RedisConnectionStateEvent.CONNECTION_DOWN, connection, connection.connmark, self))

    async def error(self, connection):
        await Protocol.error(self, connection)
        self._logger.warning('Redis connection is reset on %r', connection)
        await connection.wait_for_send(RedisConnectionStateEvent(RedisConnectionStateEvent.CONNECTION_DOWN, connection, connection.connmark, self))

    def replymatcher(self, requestid, connection, iserror = None):
        """
        Create an event matcher to match 
        """
        matcherparam = [connection, connection.connmark, requestid]
        if iserror is not None:
            matcherparam.append(iserror)
        return RedisResponseEvent.createMatcher(*matcherparam)

    def subscribematcher(self, connection, subscribe = None, channel = None, type = None):
        return RedisSubscribeMessageEvent.createMatcher(type,
                                                        self._bytes(subscribe) if subscribe is not None else None,
                                                        self._bytes(channel) if channel is not None else None, connection)

    def statematcher(self, connection, state = RedisConnectionStateEvent.CONNECTION_DOWN, currentconn = True):
        if currentconn:
            return RedisConnectionStateEvent.createMatcher(state, connection, connection.connmark)
        else:
            return RedisConnectionStateEvent.createMatcher(state, connection)

    def _prepare_command(self, connection, args):
        cmdname = _str(args[0]).upper()
        if cmdname in ('SUBSCRIBE', 'UNSUBSCRIBE', 'PSUBSCRIBE', 'PUNSUBSCRIBE'):
            if len(args) <= 1:
                raise RedisProtocolException('(P)SUBSCRIBE/(P)UNSUBSCRIBE must have at least one key as parameter')
            connection.redis_subscribe = True
            if cmdname == 'SUBSCRIBE':
                connection.redis_subscribe_keys.update(self._bytes(a) for a in args[1:])
            elif cmdname == 'UNSUBSCRIBE':
                connection.redis_subscribe_keys.difference_update(self._bytes(a) for a in args[1:])
            elif cmdname == 'PSUBSCRIBE':
                connection.redis_subscribe_pkeys.update(self._bytes(a) for a in args[1:])
            elif cmdname == 'PUNSUBSCRIBE':
                connection.redis_subscribe_pkeys.difference_update(self._bytes(a) for a in args[1:])
            reply_matcher = lambda: RedisSubscribeEvent.createMatcher(cmdname.lower(), self._bytes(args[-1]), connection, connection.connmark)
        elif connection.redis_subscribe_keys or connection.redis_subscribe_pkeys:
            if cmdname == 'PING':
                def reply_matcher():
                    rid = connection.redis_ping
                    connection.redis_ping -= 1
                    return RedisResponseEvent.createMatcher(connection, connection.connmark, rid)
            elif cmdname == 'QUIT':
                reply_matcher = lambda: RedisResponseEvent.createMatcher(connection, connection.connmark, 0)
                connection.need_reconnect = False
            else:
                raise RedisProtocolException('Can only use PING/QUIT/(P)SUBSCRIBE/(P)UNSUBSCRIBE on a subscribed connection')
        else:
            if cmdname == 'SELECT':
                if len(args) != 2:
                    raise RedisProtocolException("wrong number of arguments for 'select' command")
                connection.redis_select = _str(args[1])
            elif cmdname == 'QUIT':
                connection.need_reconnect = False
            def reply_matcher():
                rid = connection.xid
                connection.xid += 1
                return RedisResponseEvent.createMatcher(connection, connection.connmark, rid)
        r = self.format_request(*args)
        return r, reply_matcher

    async def send_command(self, connection, container, *args):
        '''
        Send command to Redis server.
        
        :param connection: Redis connection
        
        :param container: routine container
        
        :param \*args: command paramters, begin with command name, e.g. `'SET'`,`'key'`,`'value'`
        
        :returns: Event matcher to wait for reply
        '''
        if not args:
            raise RedisProtocolException('No command name')
        r, reply_matcher = self._prepare_command(connection, args)
        await connection.write(ConnectionWriteEvent(connection, connection.connmark, data = r), False)
        return reply_matcher()

    async def send_batch(self, connection, container, *cmds):
        '''
        Send multiple commands to redis server at once
        
        :param connection: redis connection
        
        :param container: routine container
        
        :param \*cmds: commands to send. Each command is a tuple/list of bytes/str.
        
        :returns: list of reply event matchers
        '''
        if not cmds:
            raise RedisProtocolException('No commands')
        commands = []
        matchers = []
        for c in cmds:
            r, reply_matcher = self._prepare_command(connection, c)
            commands.append(r)
            matchers.append(reply_matcher)
        await connection.write(ConnectionWriteEvent(connection, connection.connmark, data = b''.join(commands)), False)
        return [m() for m in matchers]

    async def execute_command(self, connection, container, *args):
        '''
        Send command to Redis server and wait for response
        
        :param connection: Redis connection
        
        :param container: routine container
        
        :param \*args: command paramters, begin with command name, e.g. `'SET'`,`'key'`,`'value'`
        
        :returns: Response from Redis server
        
        :raises RedisReplyException: Redis server returns an error (e.g. "-ERR ...")
        '''
        rm = await self.send_command(connection, container, *args)
        sm = self.statematcher(connection)
        ev, m = await M_(sm, rm)
        if m is sm:
            raise RedisProtocolException('Redis connection down before response received')
        else:
            r = ev.result
            if isinstance(r, Exception):
                raise r
            else:
                return r

    async def batch_execute(self, connection, container, *cmds, raise_first_exception = False):
        '''
        Send multiple commands to redis server at once, and get responses
        
        :param connection: redis connection
        
        :param container: routine container
        
        :param \*cmds: commands to send. Each command is a tuple/list of bytes/str.
        
        :param raise_first_exception: if True, the first exception is raised.
                                      if False, exceptions are returned in the list.
        
        :returns: list of replies.
        '''
        if not cmds:
            return []
        matchers = await self.send_batch(connection, container, *cmds)
        sm = self.statematcher(connection)
        retvalue = []
        for m in matchers:
            ev, m = await M_(m, sm)
            if m is sm:
                raise RedisProtocolException('Redis connection down before response received')
            r = ev.result
            if raise_first_exception and isinstance(r, Exception):
                raise r
            retvalue.append(r)
        return retvalue

    def parse(self, connection, data, laststart):
        events = []
        connection.redis_reader.feed(_copy(data))
        while True:
            r = connection.redis_reader.gets()
            if r is False:
                break
            if connection.redis_replyxid < connection.xid:
                events.append(RedisResponseEvent(connection, connection.connmark, connection.redis_replyxid, isinstance(r, Exception), self, result = r))
                connection.redis_replyxid += 1
            elif connection.redis_subscribe:
                if isinstance(r, bytes) or isinstance(r, Exception):
                    events.append(RedisResponseEvent(connection, connection.connmark, 0, isinstance(r, Exception), self, result = r))
                elif r[0] == b'message':
                    events.append(RedisSubscribeMessageEvent(RedisSubscribeMessageEvent.MESSAGE, r[1], r[1],
                                                             connection, connection.connmark, self, message = r[2]))
                elif r[0] == b'pmessage':
                    events.append(RedisSubscribeMessageEvent(RedisSubscribeMessageEvent.PMESSAGE, r[1], r[2],
                                                             connection, connection.connmark, self, message = r[3]))
                elif r[0] == b'pong':
                    events.append(RedisResponseEvent(connection, connection.connmark, connection.redis_pingreply, False, self, result = r[1]))
                    connection.redis_pingreply -= 1
                else:
                    events.append(RedisSubscribeEvent(_str(r[0]), r[1], connection, connection.connmark, self, result = r))
                    if not r[2]:
                        connection.redis_subscribe = False
                        connection.xid += connection.redis_bufferedxid
        if laststart == len(data):
            # Remote write close
            events.append(ConnectionWriteEvent(connection, connection.connmark, data = b'', EOF = True))
        return (events, 0)

    async def notconnected(self, connection):
        await Protocol.notconnected(self, connection)
        await connection.wait_for_send(RedisConnectionStateEvent(RedisConnectionStateEvent.CONNECTION_NOTCONNECTED,
                                                                  connection, connection.connmark, self))
    def has_pending_requests(self, connection):
        return connection.redis_replyxid != connection.xid

    async def keepalive(self, connection):
        try:
            if connection.redis_replyxid == connection.xid:
                timeout, _ = await connection.execute_with_timeout(
                                            self.keepalivetimeout,
                                            self.execute_command(connection, connection, 'PING'))
                if timeout:
                    await connection.reset(True)
        except Exception:
            await connection.reset(True)

    @staticmethod
    def reconnect_timeseq():
        yield 0
        nextSeq = 0.5
        while True:
            yield nextSeq                
            if nextSeq < 16:
                nextSeq = nextSeq * 2
            else:
                nextSeq = 20



'''
Created on 2016/2/26

:author: think
'''


from vlcp.config import defaultconfig
from vlcp.server.module import Module, api, depend, call_api, ModuleNotification
from vlcp.event.runnable import RoutineContainer
from vlcp.service.sdn import ovsdbmanager
from vlcp.event.connection import ConnectionResetException
from vlcp.event.event import Event, withIndices, M_
from vlcp.service.sdn.ovsdbmanager import OVSDBBridgeSetup, OVSDBConnectionSetup,\
    OVSDBBridgeNotAppearException
import vlcp.utils.ovsdb as ovsdb
from vlcp.protocol.jsonrpc import JsonRPCErrorResultException,\
    JsonRPCNotificationEvent, JsonRPCProtocolException
import itertools
from contextlib import closing

def _bytes(s):
    return s.encode('ascii')

@withIndices('connection')
class OVSDBConnectionPortsSynchronized(Event):
    pass

@withIndices('connection', 'name', 'ofport', 'id', 'vhost', 'datapathid')
class OVSDBPortUpNotification(Event):
    pass

class OVSDBPortNotAppearException(Exception):
    pass


@defaultconfig
@depend(ovsdbmanager.OVSDBManager)
class OVSDBPortManager(Module):
    '''
    Manage Ports from OVSDB Protocol
    '''
    service = True
    def __init__(self, server):
        Module.__init__(self, server)
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._manage_ports
        self.routines.append(self.apiroutine)
        self.managed_ports = {}
        self.managed_ids = {}
        self.monitor_routines = set()
        self.ports_uuids = {}
        self.wait_portnos = {}
        self.wait_names = {}
        self.wait_ids = {}
        self.bridge_datapathid = {}
        self.createAPI(api(self.getports, self.apiroutine),
                       api(self.getallports, self.apiroutine),
                       api(self.getportbyid, self.apiroutine),
                       api(self.waitportbyid, self.apiroutine),
                       api(self.getportbyname, self.apiroutine),
                       api(self.waitportbyname, self.apiroutine),
                       api(self.getportbyno, self.apiroutine),
                       api(self.waitportbyno, self.apiroutine),
                       api(self.resync, self.apiroutine)
                       )
        self._synchronized = False

    async def _get_interface_info(self, connection, protocol, buuid, interface_uuid, port_uuid):
        try:
            method, params = ovsdb.transact('Open_vSwitch',
                                            ovsdb.wait('Interface', [["_uuid", "==", ovsdb.uuid(interface_uuid)]],
                                                       ["ofport"], [{"ofport":ovsdb.oset()}], False, 5000),
                                            ovsdb.wait('Interface', [["_uuid", "==", ovsdb.uuid(interface_uuid)]],
                                                       ["ofport"], [{"ofport":-1}], False, 0),
                                            ovsdb.wait('Interface', [["_uuid", "==", ovsdb.uuid(interface_uuid)]],
                                                       ["ifindex"], [{"ifindex":ovsdb.oset()}], False, 5000),
                                            ovsdb.select('Interface', [["_uuid", "==", ovsdb.uuid(interface_uuid)]],
                                                                         ["_uuid", "name", "ifindex", "ofport", "type", "external_ids"]))
            result, _ = await protocol.querywithreply(method, params, connection, self.apiroutine)
            r = result[0]
            if 'error' in r:
                raise JsonRPCErrorResultException('Error while acquiring interface: ' + repr(r['error']))            
            r = result[1]
            if 'error' in r:
                raise JsonRPCErrorResultException('Error while acquiring interface: ' + repr(r['error']))            
            r = result[2]
            if 'error' in r:
                # Ignore this port because it is in an error state
                return []
            r = result[3]
            if 'error' in r:
                raise JsonRPCErrorResultException('Error while acquiring interface: ' + repr(r['error']))
            if not r['rows']:
                return []
            r0 = r['rows'][0]
            if r0['ofport'] < 0:
                # Ignore this port because it is in an error state
                return []
            r0['_uuid'] = r0['_uuid'][1]
            r0['ifindex'] = ovsdb.getoptional(r0['ifindex'])
            r0['external_ids'] = ovsdb.getdict(r0['external_ids'])
            if buuid not in self.bridge_datapathid:
                return []
            else:
                datapath_id = self.bridge_datapathid[buuid]
            if 'iface-id' in r0['external_ids']:
                eid = r0['external_ids']['iface-id']
                r0['id'] = eid
                id_ports = self.managed_ids.setdefault((protocol.vhost, eid), [])
                id_ports.append((datapath_id, r0))
            else:
                r0['id'] = None
            self.managed_ports.setdefault((protocol.vhost, datapath_id),[]).append((port_uuid, r0))
            notify = False
            if (protocol.vhost, datapath_id, r0['ofport']) in self.wait_portnos:
                notify = True
                del self.wait_portnos[(protocol.vhost, datapath_id, r0['ofport'])]
            if (protocol.vhost, datapath_id, r0['name']) in self.wait_names:
                notify = True
                del self.wait_names[(protocol.vhost, datapath_id, r0['name'])]
            if (protocol.vhost, r0['id']) in self.wait_ids:
                notify = True
                del self.wait_ids[(protocol.vhost, r0['id'])]
            if notify:
                await self.apiroutine.wait_for_send(OVSDBPortUpNotification(connection, r0['name'],
                                                                             r0['ofport'], r0['id'],
                                                                             protocol.vhost, datapath_id,
                                                                             port = r0))
            return [r0]
        except JsonRPCProtocolException:
            return []

    def _remove_interface_id(self, connection, protocol, datapath_id, port):
        eid = port['id']
        eid_list = self.managed_ids.get((protocol.vhost, eid))
        for i in range(0, len(eid_list)):
            if eid_list[i][1]['_uuid'] == port['_uuid']:
                del eid_list[i]
                break

    def _remove_interface(self, connection, protocol, datapath_id, interface_uuid, port_uuid):
        ports = self.managed_ports.get((protocol.vhost, datapath_id))
        r = None
        if ports is not None:
            for i in range(0, len(ports)):
                if ports[i][1]['_uuid'] == interface_uuid:
                    r = ports[i][1]
                    if r['id']:
                        self._remove_interface_id(connection, protocol, datapath_id, r)
                    del ports[i]
                    break
            if not ports:
                del self.managed_ports[(protocol.vhost, datapath_id)]
        return r

    def _remove_all_interface(self, connection, protocol, datapath_id, port_uuid, buuid):
        ports = self.managed_ports.get((protocol.vhost, datapath_id))
        if ports is not None:
            removed_ports = [r for puuid, r in ports if puuid == port_uuid]
            not_removed_ports = [(puuid, r) for puuid, r in ports if puuid != port_uuid]
            ports[:len(not_removed_ports)] = not_removed_ports
            del ports[len(not_removed_ports):]
            for r in removed_ports:
                if r['id']:
                    self._remove_interface_id(connection, protocol, datapath_id, r)
            if not ports:
                del self.managed_ports[(protocol.vhost, datapath_id)]
            return removed_ports
        if port_uuid in self.ports_uuids and self.ports_uuids[port_uuid] == buuid:
            del self.ports_uuids[port_uuid]
        return []

    async def _update_interfaces(self, connection, protocol, updateinfo, update = True):
        """
        There are several kinds of updates, they may appear together:
        
        1. New bridge created (or from initial updateinfo). We should add all the interfaces to the list.
        
        2. Bridge removed. Remove all the ports.
        
        3. name and datapath_id may be changed. We will consider this as a new bridge created, and an old
           bridge removed.
        
        4. Bridge ports modification, i.e. add/remove ports.
           a) Normally a port record is created/deleted together. A port record cannot exist without a
              bridge containing it.
               
           b) It is also possible that a port is removed from one bridge and added to another bridge, in
              this case the ports do not appear in the updateinfo
            
        5. Port interfaces modification, i.e. add/remove interfaces. The bridge record may not appear in this
           situation.
           
        We must consider these situations carefully and process them in correct order.
        """
        port_update = updateinfo.get('Port', {})
        bridge_update = updateinfo.get('Bridge', {})
        working_routines = []
        async def process_bridge(buuid, uo):
            try:
                nv = uo['new']
                if 'datapath_id' in nv:
                    if ovsdb.getoptional(nv['datapath_id']) is None:
                        # This bridge is not initialized. Wait for the bridge to be initialized.
                        datapath_id = await call_api(self.apiroutine, 'ovsdbmanager', 'waitbridge',
                                                     {'connection': connection,
                                                      'name': nv['name'],
                                                      'timeout': 5})
                    else:
                        datapath_id = int(nv['datapath_id'], 16)
                    self.bridge_datapathid[buuid] = datapath_id
                elif buuid in self.bridge_datapathid:
                    datapath_id = self.bridge_datapathid[buuid]
                else:
                    # This should not happen, but just in case...
                    datapath_id = await call_api(self.apiroutine, 'ovsdbmanager', 'waitbridge',
                                                                {'connection': connection,
                                                                 'name': nv['name'],
                                                                 'timeout': 5})
                    self.bridge_datapathid[buuid] = datapath_id
                if 'ports' in nv:
                    nset = set((p for _,p in ovsdb.getlist(nv['ports'])))
                else:
                    nset = set()
                if 'old' in uo:
                    ov = uo['old']
                    if 'ports' in ov:
                        oset = set((p for _,p in ovsdb.getlist(ov['ports'])))
                    else:
                        # new ports are not really added; it is only sent because datapath_id is modified
                        nset = set()
                        oset = set()
                    if 'datapath_id' in ov and ovsdb.getoptional(ov['datapath_id']) is not None:
                        old_datapathid = int(ov['datapath_id'], 16)
                    else:
                        old_datapathid = datapath_id
                else:
                    oset = set()
                    old_datapathid = datapath_id
                # For every deleted port, remove the interfaces with this port _uuid
                remove = []
                add_routine = []
                for puuid in oset - nset:
                    remove += self._remove_all_interface(connection, protocol, old_datapathid, puuid, buuid)
                # For every port not changed, check if the interfaces are modified;
                for puuid in oset.intersection(nset):
                    if puuid in port_update:
                        # The port is modified, there should be an 'old' set and 'new' set
                        pu = port_update[puuid]
                        if 'old' in pu:
                            poset = set((p for _,p in ovsdb.getlist(pu['old']['interfaces'])))
                        else:
                            poset = set()
                        if 'new' in pu:
                            pnset = set((p for _,p in ovsdb.getlist(pu['new']['interfaces'])))
                        else:
                            pnset = set()
                        # Remove old interfaces
                        remove += [r for r in 
                                   (self._remove_interface(connection, protocol, datapath_id, iuuid, puuid)
                                    for iuuid in (poset - pnset)) if r is not None]
                        # Prepare to add new interfaces
                        add_routine += [self._get_interface_info(connection, protocol, buuid, iuuid, puuid)
                                        for iuuid in (pnset - poset)]
                # For every port added, add the interfaces
                async def add_port_interfaces(puuid):
                    # If the uuid does not appear in update info, we have no choice but to query interfaces with select
                    # we cannot use data from other bridges; the port may be moved from a bridge which is not tracked
                    try:
                        method, params = ovsdb.transact('Open_vSwitch', ovsdb.select('Port',
                                                                                     [["_uuid", "==", ovsdb.uuid(puuid)]],
                                                                                     ["interfaces"]))
                        result, _ = await protocol.querywithreply(method, params, connection, self.apiroutine)
                        r = result[0]
                        if 'error' in r:
                            raise JsonRPCErrorResultException('Error when query interfaces from port ' + repr(puuid) + ': ' + r['error'])
                        if r['rows']:
                            interfaces = ovsdb.getlist(r['rows'][0]['interfaces'])
                            result = await self.apiroutine.execute_all([self._get_interface_info(connection, protocol, buuid, iuuid, puuid)
                                                                        for _,iuuid in interfaces])
                            return list(itertools.chain(r[0] for r in result))
                        else:
                            return []
                    except JsonRPCProtocolException:
                        return []
                    except ConnectionResetException:
                        return []

                for puuid in nset - oset:
                    self.ports_uuids[puuid] = buuid
                    if puuid in port_update and 'new' in port_update[puuid] \
                            and 'old' not in port_update[puuid]:
                        # Add all the interfaces in 'new'
                        interfaces = ovsdb.getlist(port_update[puuid]['new']['interfaces'])
                        add_routine += [self._get_interface_info(connection, protocol, buuid, iuuid, puuid)
                                        for _,iuuid in interfaces]
                    else:
                        add_routine.append(add_port_interfaces(puuid))
                # Execute the add_routine
                try:
                    result = await self.apiroutine.execute_all(add_routine)
                except:
                    add = []
                    raise
                else:
                    add = list(itertools.chain(r[0] for r in result))
                finally:
                    if update:
                        self.scheduler.emergesend(
                                ModuleNotification(self.getServiceName(), 'update',
                                                   datapathid = datapath_id,
                                                   connection = connection,
                                                   vhost = protocol.vhost,
                                                   add = add, remove = remove,
                                                   reason = 'bridgemodify'
                                                            if 'old' in uo
                                                            else 'bridgeup'
                                                   ))
            except JsonRPCProtocolException:
                pass
            except ConnectionResetException:
                pass
            except OVSDBBridgeNotAppearException:
                pass
        ignore_ports = set()
        for buuid, uo in bridge_update.items():
            # Bridge removals are ignored because we process OVSDBBridgeSetup event instead
            if 'old' in uo:
                if 'ports' in uo['old']:
                    oset = set((puuid for _, puuid in ovsdb.getlist(uo['old']['ports'])))
                    ignore_ports.update(oset)
                if 'new' not in uo:
                    if buuid in self.bridge_datapathid:
                        del self.bridge_datapathid[buuid]
            if 'new' in uo:
                # If bridge contains this port is updated, we process the port update totally in bridge,
                # so we ignore it later
                if 'ports' in uo['new']:
                    nset = set((puuid for _, puuid in ovsdb.getlist(uo['new']['ports'])))
                    ignore_ports.update(nset)
                working_routines.append(process_bridge(buuid, uo))

        async def process_port(buuid, port_uuid, interfaces, remove_ids):
            if buuid not in self.bridge_datapathid:
                return
            datapath_id = self.bridge_datapathid[buuid]
            ports = self.managed_ports.get((protocol.vhost, datapath_id))
            remove = []
            if ports is not None:
                remove = [p for _,p in ports if p['_uuid'] in remove_ids]
                not_remove = [(_,p) for _,p in ports if p['_uuid'] not in remove_ids]
                ports[:len(not_remove)] = not_remove
                del ports[len(not_remove):]
            if interfaces:
                try:
                    result = await self.apiroutine.execute_all(
                                    [self._get_interface_info(connection, protocol, buuid, iuuid, port_uuid)
                                     for iuuid in interfaces])
                    add = list(itertools.chain((r[0] for r in result if r[0])))
                except Exception:
                    self._logger.warning("Cannot get new port information", exc_info = True)
                    add = []
            else:
                add = []
            if update:
                await self.apiroutine.wait_for_send(ModuleNotification(self.getServiceName(), 'update', datapathid = datapath_id,
                                                                                                      connection = connection,
                                                                                                      vhost = protocol.vhost,
                                                                                                      add = add, remove = remove,
                                                                                                      reason = 'bridgemodify'
                                                                                                      ))
        for puuid, po in port_update.items():
            if puuid not in ignore_ports:
                bridge_id = self.ports_uuids.get(puuid)
                if bridge_id is not None:
                    datapath_id = self.bridge_datapathid[bridge_id]
                    if datapath_id is not None:
                        # This port is modified
                        if 'new' in po:
                            nset = set((iuuid for _, iuuid in ovsdb.getlist(po['new']['interfaces'])))
                        else:
                            nset = set()                    
                        if 'old' in po:
                            oset = set((iuuid for _, iuuid in ovsdb.getlist(po['old']['interfaces'])))
                        else:
                            oset = set()
                        working_routines.append(process_port(bridge_id, puuid, nset - oset, oset - nset))
        if update:
            for r in working_routines:
                self.apiroutine.subroutine(r)
        else:
            try:
                await self.apiroutine.execute_all(working_routines)
            finally:
                self.scheduler.emergesend(OVSDBConnectionPortsSynchronized(connection))

    async def _get_ports(self, connection, protocol):
        try:
            try:
                method, params = ovsdb.monitor('Open_vSwitch', 'ovsdb_port_manager_interfaces_monitor', {
                                                    'Bridge':[ovsdb.monitor_request(["name", "datapath_id", "ports"])],
                                                    'Port':[ovsdb.monitor_request(["interfaces"])]
                                                })
                try:
                    r, _ = await protocol.querywithreply(method, params, connection, self.apiroutine)
                except JsonRPCErrorResultException:
                    # The monitor is already set, cancel it first
                    method2, params2 = ovsdb.monitor_cancel('ovsdb_port_manager_interfaces_monitor')
                    await protocol.querywithreply(method2, params2, connection, self.apiroutine, False)
                    r, _ = await protocol.querywithreply(method, params, connection, self.apiroutine)
            except:
                async def _msg():
                    await self.apiroutine.wait_for_send(OVSDBConnectionPortsSynchronized(connection))
                self.apiroutine.subroutine(_msg(), False)
                raise
            # This is the initial state, it should contains all the ids of ports and interfaces
            self.apiroutine.subroutine(self._update_interfaces(connection, protocol, r, False))
            update_matcher = JsonRPCNotificationEvent.createMatcher('update', connection, connection.connmark,
                                                                    _ismatch = lambda x: x.params[0] == 'ovsdb_port_manager_interfaces_monitor')
            conn_state = protocol.statematcher(connection)
            while True:
                ev, m = await M_(update_matcher, conn_state)
                if m is conn_state:
                    break
                else:
                    self.apiroutine.subroutine(self._update_interfaces(connection, protocol, ev.params[1], True))
        except JsonRPCProtocolException:
            pass
        finally:
            if self.apiroutine.currentroutine in self.monitor_routines:
                self.monitor_routines.remove(self.apiroutine.currentroutine)

    async def _get_existing_ports(self):
        r = await call_api(self.apiroutine, 'ovsdbmanager', 'getallconnections', {'vhost':None})
        matchers = []
        for c in r:
            self.monitor_routines.add(self.apiroutine.subroutine(self._get_ports(c, c.protocol)))
            matchers.append(OVSDBConnectionPortsSynchronized.createMatcher(c))
        await self.apiroutine.wait_for_all(*matchers)
        self._synchronized = True
        await self.apiroutine.wait_for_send(ModuleNotification(self.getServiceName(), 'synchronized'))

    async def _wait_for_sync(self):
        if not self._synchronized:
            await ModuleNotification.createMatcher(self.getServiceName(), 'synchronized')

    async def _manage_ports(self):
        try:
            self.apiroutine.subroutine(self._get_existing_ports())
            connsetup = OVSDBConnectionSetup.createMatcher()
            bridgedown = OVSDBBridgeSetup.createMatcher(OVSDBBridgeSetup.DOWN)
            while True:
                e, m = await M_(connsetup, bridgedown)
                if m is connsetup:
                    self.monitor_routines.add(self.apiroutine.subroutine(self._get_ports(e.connection, e.connection.protocol)))
                else:
                    # Remove ports of the bridge
                    ports =  self.managed_ports.get((e.vhost, e.datapathid))
                    if ports is not None:
                        ports_original = ports
                        ports = [p for _,p in ports]
                        for p in ports:
                            if p['id']:
                                self._remove_interface_id(e.connection,
                                                          e.connection.protocol, e.datapathid, p)
                        newdpid = getattr(e, 'new_datapath_id', None)
                        buuid = e.bridgeuuid
                        if newdpid is not None:
                            # This bridge changes its datapath id
                            if buuid in self.bridge_datapathid and self.bridge_datapathid[buuid] == e.datapathid:
                                self.bridge_datapathid[buuid] = newdpid
                            async def re_add_interfaces():
                                result = await self.apiroutine.execute_all(
                                                [self._get_interface_info(e.connection, e.connection.protocol, buuid,
                                                                          r['_uuid'], puuid)
                                                 for puuid, r in ports_original])
                                add = list(itertools.chain(r[0] for r in result))
                                await self.apiroutine.wait_for_send(ModuleNotification(self.getServiceName(),
                                                  'update', datapathid = e.datapathid,
                                                  connection = e.connection,
                                                  vhost = e.vhost,
                                                  add = add, remove = [],
                                                  reason = 'bridgeup'
                                                  ))
                            self.apiroutine.subroutine(re_add_interfaces())
                        else:
                            # The ports are removed
                            for puuid, _ in ports_original:
                                if puuid in self.ports_uuids[puuid] and self.ports_uuids[puuid] == buuid:
                                    del self.ports_uuids[puuid]
                        del self.managed_ports[(e.vhost, e.datapathid)]
                        self.scheduler.emergesend(ModuleNotification(self.getServiceName(),
                                                  'update', datapathid = e.datapathid,
                                                  connection = e.connection,
                                                  vhost = e.vhost,
                                                  add = [], remove = ports,
                                                  reason = 'bridgedown'
                                                  ))
        finally:
            for r in list(self.monitor_routines):
                r.close()
            self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'unsynchronized'))

    async def getports(self, datapathid, vhost = ''):
        "Return all ports of a specifed datapath"
        await self._wait_for_sync()
        return [p for _,p in self.managed_ports.get((vhost, datapathid), [])]

    async def getallports(self, vhost = None):
        "Return all ``(datapathid, port, vhost)`` tuples, optionally filterd by vhost"
        await self._wait_for_sync()
        if vhost is None:
            return [(dpid, p, vh) for (vh, dpid),v in self.managed_ports.items() for _,p in v]
        else:
            return [(dpid, p, vh) for (vh, dpid),v in self.managed_ports.items() if vh == vhost for _,p in v]

    async def getportbyno(self, datapathid, portno, vhost = ''):
        "Return port with specified portno"
        portno &= 0xffff
        await self._wait_for_sync()
        return self._getportbyno(datapathid, portno, vhost)

    def _getportbyno(self, datapathid, portno, vhost = ''):
        ports = self.managed_ports.get((vhost, datapathid))
        if ports is None:
            return None
        else:
            for _, p in ports:
                if p['ofport'] == portno:
                    return p
            return None

    async def waitportbyno(self, datapathid, portno, timeout = 30, vhost = ''):
        "Wait for port with specified portno"
        portno &= 0xffff
        await self._wait_for_sync()
        async def waitinner():
            p = self._getportbyno(datapathid, portno, vhost)
            if p is not None:
                return p
            else:
                try:
                    self.wait_portnos[(vhost, datapathid, portno)] = \
                            self.wait_portnos.get((vhost, datapathid, portno),0) + 1
                    ev = await OVSDBPortUpNotification.createMatcher(None, None, portno, None, vhost, datapathid)
                except:
                    v = self.wait_portnos.get((vhost, datapathid, portno))
                    if v is not None:
                        if v <= 1:
                            del self.wait_portnos[(vhost, datapathid, portno)]
                        else:
                            self.wait_portnos[(vhost, datapathid, portno)] = v - 1
                    raise
                else:
                    return ev.port
        timeout_, r = await self.apiroutine.execute_with_timeout(timeout, waitinner())
        if timeout_:
            raise OVSDBPortNotAppearException('Port ' + repr(portno) + ' does not appear before timeout')
        else:
            return r

    async def getportbyname(self, datapathid, name, vhost = ''):
        "Return port with specified name"
        if isinstance(name, bytes):
            name = name.decode('utf-8')
        await self._wait_for_sync()
        return self._getportbyname(datapathid, name, vhost)

    def _getportbyname(self, datapathid, name, vhost = ''):
        ports = self.managed_ports.get((vhost, datapathid))
        if ports is None:
            return None
        else:
            for _, p in ports:
                if p['name'] == name:
                    return p
            return None

    async def waitportbyname(self, datapathid, name, timeout = 30, vhost = ''):
        "Wait for port with specified name"
        await self._wait_for_sync()
        async def waitinner():
            p = self._getportbyname(datapathid, name, vhost)
            if p is not None:
                return p
            else:
                try:
                    self.wait_names[(vhost, datapathid, name)] = \
                            self.wait_portnos.get((vhost, datapathid, name) ,0) + 1
                    ev = await OVSDBPortUpNotification.createMatcher(None, name, None, None, vhost, datapathid)
                except:
                    v = self.wait_names.get((vhost, datapathid, name))
                    if v is not None:
                        if v <= 1:
                            del self.wait_names[(vhost, datapathid, name)]
                        else:
                            self.wait_names[(vhost, datapathid, name)] = v - 1
                    raise
                else:
                    return ev.port
        timeout_, r = await self.apiroutine.execute_with_timeout(timeout, waitinner())
        if timeout_:
            raise OVSDBPortNotAppearException('Port ' + repr(name) + ' does not appear before timeout')
        else:
            return r

    async def getportbyid(self, id, vhost = ''):
        "Return port with the specified id. The return value is a pair: ``(datapath_id, port)``"
        await self._wait_for_sync()
        return self._getportbyid(id, vhost)

    def _getportbyid(self, id, vhost = ''):
        ports = self.managed_ids.get((vhost, id))
        if ports:
            return ports[0]
        else:
            return None

    async def waitportbyid(self, id, timeout = 30, vhost = ''):
        "Wait for port with the specified id. The return value is a pair ``(datapath_id, port)``"
        await self._wait_for_sync()
        async def waitinner():
            p = self._getportbyid(id, vhost)
            if p is None:
                try:
                    self.wait_ids[(vhost, id)] = self.wait_ids.get((vhost, id), 0) + 1
                    ev = await OVSDBPortUpNotification.createMatcher(None, None, None, id, vhost)
                except:
                    v = self.wait_ids.get((vhost, id))
                    if v is not None:
                        if v <= 1:
                            del self.wait_ids[(vhost, id)]
                        else:
                            self.wait_ids[(vhost, id)] = v - 1
                    raise
                else:
                    return (ev.datapathid, ev.port)
            else:
                return p
        timeout_, r = await self.apiroutine.execute_with_timeout(timeout, waitinner())
        if timeout_:
            raise OVSDBPortNotAppearException('Port ' + repr(id) + ' does not appear before timeout')
        else:
            return r
    
    async def resync(self, datapathid, vhost = ''):
        '''
        Resync with current ports
        '''
        # Sometimes when the OVSDB connection is very busy, monitor message may be dropped.
        # We must deal with this and recover from it
        # Save current manged_ports
        if (vhost, datapathid) not in self.managed_ports:
            return
        else:
            c = await call_api(self.apiroutine, 'ovsdbmanager', 'getconnection', {'datapathid': datapathid, 'vhost':vhost})
            if c is not None:
                # For now, we restart the connection...
                await c.reconnect(False)
                await self.apiroutine.wait_with_timeout(0.1)
                await call_api(self.apiroutine, 'ovsdbmanager', 'waitconnection', {'datapathid': datapathid,
                                                                                     'vhost': vhost})



'''
Created on 2018/7/2

:author: hubo

Base walkers for network plugins
'''
from vlcp.utils.dataobject import create_new
from vlcp.utils.networkmodel import PhysicalNetwork, PhysicalNetworkMap,\
    PhysicalNetworkSet, PhysicalPort, PhysicalPortSet, LogicalNetworkSet,\
    LogicalNetwork, LogicalNetworkMap
from functools import partial


def default_processor(obj, *args, parameters, excluding=('id',), disabled=()):
    _updated = False
    for k,v in parameters.items():
        if k in disabled:
            raise ValueError(repr(k) + " cannot be updated")
        if k not in excluding:
            _updated = True
            setattr(obj, k, v)
    return _updated


def _false_processor(*args, **kwargs):
    return False


def default_physicalnetwork_keys(id_, parameters):
    return (PhysicalNetworkMap.default_key(id_),
            PhysicalNetworkSet.default_key())


def default_physicalnetwork_update_keys(id_, parameters):
    return (PhysicalNetworkMap.default_key(id_),)


def default_iterate_dict(parameter_dict):
    return parameter_dict.items()


def createphysicalnetwork(type, create_processor = partial(default_processor, excluding=('id', 'type')),
                                      reorder_dict = default_iterate_dict):
    """
    :param type: physical network type
    
    :param create_processor: create_processor(physicalnetwork, walk, write, \*, parameters)
    """
    # create an new physical network
    def walker(walk, write, timestamp, parameters_dict):
        for key, parameters in reorder_dict(parameters_dict):
            try:
                value = walk(key)
            except KeyError:
                pass
            else:
                id_ = parameters['id']
                new_network = create_new(PhysicalNetwork, value, id_)
                new_network.type = type
                
                create_processor(new_network, walk, write, parameters=parameters)
                write(key, new_network)
                new_networkmap = PhysicalNetworkMap.create_instance(id_)
                new_networkmap.network = new_network.create_weakreference()
                write(new_networkmap.getkey(), new_networkmap)
                
                # Save into network set
                try:
                    physet = walk(PhysicalNetworkSet.default_key())
                except KeyError:
                    pass
                else:
                    physet.set.dataset().add(new_network.create_weakreference())
                    write(physet.getkey(), physet)
    return walker


def updatephysicalnetwork(update_processor = partial(default_processor, disabled=('type',)),
                                reorder_dict = default_iterate_dict):
    """
    :param update_processor: update_processor(physicalnetwork, walk, write, \*, parameters)
    """
    def walker(walk, write, timestamp, parameters_dict):
        # do not need to check value is not None in a plugin; it is checked in the caller
        for key, parameters in reorder_dict(parameters_dict):
            try:
                value = walk(key)
            except KeyError:
                pass
            else:
                if update_processor(value, walk, write, parameters=parameters):
                    write(key, value)
    return walker


def default_physicalnetwork_delete_check(phynet, phymap, *args, parameters):
    if phymap.logicnetworks.dataset():
        raise ValueError("Must delete all logical networks of this physical network "+ parameters['id'])
    if phymap.ports.dataset():
        raise ValueError("Must delete all physical ports of this physical network " + parameters['id'])


def deletephysicalnetwork(check_processor = default_physicalnetwork_delete_check,
                                reorder_dict = default_iterate_dict):
    """
    :param check_processor: check_processor(physicalnetwork, physicalnetworkmap, walk, write, \*, parameters)
    """
    def walker(walk, write, timestamp, parameters_dict):
        for key, parameters in reorder_dict(parameters_dict):
            try:
                value = walk(key)
            except KeyError:
                pass
            else:
                id_ = parameters['id']
                try:
                    phy_map = walk(PhysicalNetworkMap.default_key(id_))
                except KeyError:
                    pass
                else:
                    check_processor(value, phy_map, walk, write, parameters=parameters)
                    write(phy_map.getkey(), None)
                try:
                    phynetset = walk(PhysicalNetworkSet.default_key())
                except KeyError:
                    pass
                else:
                    phynetset.set.dataset().discard(value.create_weakreference())
                    write(phynetset.getkey(), phynetset)
                write(key, None)
    return walker


def default_physicalport_keys(phynet_id, parameters):
    return (PhysicalNetwork.default_key(phynet_id),
            PhysicalNetworkMap.default_key(phynet_id),
            PhysicalPortSet.default_key())


def createphysicalport(create_processor = partial(default_processor, excluding=('vhost', 'systemid',
                                                                                       'bridge', 'name',
                                                                                       'physicalnetwork')),
                       reorder_dict = default_iterate_dict):
    """
    :param create_processor: create_processor(physicalport, physicalnetwork, physicalnetworkmap, walk, write, \*, parameters)
    """
    def walker(walk, write, timestamp, parameters_dict):
        for key, parameters in reorder_dict(parameters_dict):
            try:
                value = walk(key)
            except KeyError:
                pass
            else:
                p = create_new(PhysicalPort, value, parameters['vhost'], parameters['systemid'],
                                                    parameters['bridge'], parameters['name'])
                try:
                    physicalnetwork = walk(PhysicalNetwork.default_key(parameters['physicalnetwork']))
                except KeyError:
                    pass
                else:
                    # Should already been check from outside
                    p.physicalnetwork = physicalnetwork.create_reference()
                    try:
                        phymap = walk(PhysicalNetworkMap._network.leftkey(physicalnetwork))
                    except KeyError:
                        pass
                    else:
                        create_processor(p, physicalnetwork, phymap, walk, write, parameters=parameters)
                        phymap.ports.dataset().add(p.create_weakreference())
                        write(phymap.getkey(), phymap)
                try:
                    phyportset = walk(PhysicalPortSet.default_key())
                except KeyError:
                    pass
                else:
                    phyportset.set.dataset().add(p.create_weakreference())
                    write(phyportset.getkey(), phyportset)
                write(p.getkey(), p)
    return walker


def updatephysicalport(update_processor = partial(default_processor, excluding=('vhost', 'systemid',
                                                                                   'bridge', 'name'),
                                                                     disabled=('physicalnetwork',)),
                       reorder_dict = default_iterate_dict
                      ):
    """
    :param update_processor: update_processor(physcialport, walk, write, \*, parameters)    
    """
    def walker(walk, write, timestamp, parameters_dict):
        for key, parameters in reorder_dict(parameters_dict):
            try:
                value = walk(key)
            except KeyError:
                pass
            else:
                if update_processor(value, walk, write, parameters=parameters):
                    write(key, value)
    return walker


def deletephysicalport(check_processor=_false_processor,
                       reorder_dict = default_iterate_dict):
    """
    :param check_processor: check_processor(physicalport, physicalnetwork, physicalnetworkmap,
                            walk, write \*, parameters)
    """
    def walker(walk, write, timestamp, parameters_dict):
        for key, parameters in reorder_dict(parameters_dict):
            try:
                value = walk(key)
            except KeyError:
                pass
            else:
                try:
                    phynet = walk(value.physicalnetwork.getkey())
                except KeyError:
                    pass
                else:
                    try:
                        phymap = walk(PhysicalNetworkMap._network.leftkey(phynet))
                    except KeyError:
                        pass
                    else:
                        check_processor(value, phynet, phymap, walk, write, parameters=parameters)
                        phymap.ports.dataset().discard(value.create_weakreference())
                        write(phymap.getkey(), phymap)
                try:
                    physet = walk(PhysicalPortSet.default_key())
                except KeyError:
                    pass
                else:
                    physet.set.dataset().discard(value.create_weakreference())
                    write(physet.getkey(), physet)
                write(key, None)
    return walker


def default_logicalnetwork_keys(phynet_id, parameters):
    return (PhysicalNetwork.default_key(phynet_id),
            PhysicalNetworkMap.default_key(phynet_id),
            LogicalNetworkSet.default_key(),
            LogicalNetworkMap.default_key(parameters['id']))


def createlogicalnetwork(create_processor = partial(default_processor, excluding=('id', 'physicalnetwork')),
                         reorder_dict = default_iterate_dict):
    """
    :param create_processor: create_processor(logicalnetwork, logicalnetworkmap, physicalnetwork,
                             physicalnetworkmap, walk, write, \*, parameters)
    """
    def walker(walk, write, timestamp, parameters_dict):
        for key, parameters in reorder_dict(parameters_dict):
            try:
                value = walk(key)
            except KeyError:
                pass
            else:
                id_ = parameters['id']
                lognet = create_new(LogicalNetwork, value, id_)
                logmap = LogicalNetworkMap.create_instance(id_)
                logmap.network = lognet.create_reference()
                try:
                    phynet = walk(PhysicalNetwork.default_key(parameters['physicalnetwork']))
                except KeyError:
                    pass
                else:
                    lognet.physicalnetwork = phynet.create_reference()
                    try:
                        phymap = walk(PhysicalNetworkMap._network.leftkey(phynet))
                    except KeyError:
                        pass
                    else:
                        create_processor(lognet, logmap, phynet, phymap, walk, write, parameters=parameters)
                        phymap.logicnetworks.dataset().add(lognet.create_weakreference())
                        write(phymap.getkey(), phymap)
                        write(lognet.getkey(), lognet)
                        write(logmap.getkey(), logmap)
                try:
                    logicalnetworkset = walk(LogicalNetworkSet.default_key())
                except KeyError:
                    pass
                else:
                    logicalnetworkset.set.dataset().add(lognet.create_weakreference())
                    write(logicalnetworkset.getkey(), logicalnetworkset)
    return walker


def updatelogicalnetwork(update_processor = partial(default_processor, excluding=('id',),
                                                                       disabled=('physicalnetwork',)),
                         reorder_dict = default_iterate_dict):
    """
    :param update_processor: update_processor(logicalnetwork, walk, write, \*, parameters)
    """
    def walker(walk, write, timestamp, parameters_dict):
        for key, parameters in reorder_dict(parameters_dict):
            try:
                value = walk(key)
            except KeyError:
                pass
            else:
                if update_processor(value, walk, write, parameters=parameters):
                    write(key, value)
    return walker


def default_logicalnetwork_delete_check(logicalnetwork, logicalnetworkmap, *args, parameters):
    if logicalnetworkmap.subnets.dataset():
        raise ValueError("Must delete all subnets of this logical network "+ logicalnetwork.id)
    if logicalnetworkmap.ports.dataset():
        raise ValueError("Must delete all logical ports of this logical network " + logicalnetwork.id)


def deletelogicalnetwork(check_processor=default_logicalnetwork_delete_check,
                         reorder_dict = default_iterate_dict):
    """
    :param check_processor: check_processor(logicalnetwork, logicalnetworkmap,
                                            physicalnetwork, physicalnetworkmap,
                                            walk, write, \*, parameters)
    """
    def walker(walk, write, timestamp, parameters_dict):
        for key, parameters in reorder_dict(parameters_dict):
            try:
                value = walk(key)
            except KeyError:
                pass
            else:
                try:
                    logmap = walk(LogicalNetworkMap._network.leftkey(key))
                except KeyError:
                    pass
                else:
                    try:
                        phynet = walk(value.physicalnetwork.getkey())
                    except KeyError:
                        pass
                    else:
                        try:
                            phymap = walk(PhysicalNetworkMap._network.leftkey(phynet))
                        except KeyError:
                            pass
                        else:
                            check_processor(value, logmap, phynet, phymap, walk, write, parameters=parameters)
                            phymap.logicnetworks.dataset().discard(value.create_weakreference())
                            write(phymap.getkey(), phymap)
                            write(key, None)
                            write(logmap.getkey(), None)
                try:
                    logicalnetworkset = walk(LogicalNetworkSet.default_key())
                except KeyError:
                    pass
                else:
                    logicalnetworkset.set.dataset().discard(value.create_weakreference())
                    write(logicalnetworkset.getkey(), logicalnetworkset)
    return walker



'''
Created on 2018/7/13

:author: hubo
'''
from pychecktype import extra, TypeMismatchException
from functools import partial, wraps
from vlcp.utils.netutils import format_ip_address, format_network_cidr
from vlcp.utils.ethernet import mac_addr


def _type_assert(convert, type_, info = None):
    @wraps(convert)
    def _convert(value):
        try:
            return convert(value)
        except Exception as e:
            if info is None:
                raise TypeMismatchException(value, type_, str(e))
            else:
                raise TypeMismatchException(value, type_, info)
    return _convert


ip_address_type = extra()
ip_address_type.bind(str, convert=_type_assert(format_ip_address, ip_address_type))

cidr_type = extra()
cidr_type.bind(str, convert=_type_assert(partial(format_network_cidr, strict=True), cidr_type))

cidr_nonstrict_type = extra()
cidr_nonstrict_type.bind(str, convert=_type_assert(format_network_cidr, cidr_nonstrict_type))

mac_address_type = extra()
mac_address_type.bind(str, convert=_type_assert(lambda x: mac_addr.formatter(mac_addr(x)),
                                                mac_address_type,
                                                "Invalid MAC address"))

# Auto convert to number when calling from Web API
def _convert_to_int(s):
    try:
        return int(s)
    except ValueError:
        raise TypeMismatchException(s, autoint, "not an integer")

str_int = extra()
str_int.bind(str, convert=_type_assert(int, str_int, "not an integer"))

autoint = (int, str_int)



'''
Created on 2016/6/22

:author: hubo

See
https://tools.ietf.org/html/rfc2131
https://tools.ietf.org/html/rfc2132
'''


from namedstruct import *
from vlcp.utils.ethernet import ip4_addr, ip4_addr_bytes, mac_addr_bytes
from vlcp.utils.netutils import parse_ip4_network
from pychecktype import extra, TypeMismatchException, tuple_

boot_op = enum('boot_op', globals(), uint8,
               BOOTREQUEST = 1,
               BOOTREPLY = 2)

BOOTP_MAGIC_COOKIE = 0x63825363

dhcp_tag = enum('dhcp_tag', globals(), uint8,
                OPTION_PAD = 0,
                OPTION_END = 255,
                OPTION_NETMASK = 1,
                OPTION_TIME_OFFSET = 2,
                OPTION_ROUTER = 3,
                OPTION_DNSSERVER = 6,
                OPTION_RLSERVER = 11,
                OPTION_HOSTNAME = 12,
                # Ignore many other options...
                OPTION_DOMAINNAME = 15,
                OPTION_MTU = 26,
                OPTION_BROADCAST = 28,
                OPTION_NTPSERVER = 42,
                OPTION_REQUESTED_IP = 50,
                OPTION_LEASE_TIME = 51,
                OPTION_OVERLOAD = 52,
                OPTION_MESSAGE_TYPE = 53,
                OPTION_SERVER_IDENTIFIER = 54,
                OPTION_REQUESTED_OPTIONS = 55,
                OPTION_MESSAGE = 56,
                OPTION_MAX_MESSAGE_SIZE = 57,
                OPTION_T1 = 58,
                OPTION_T2 = 59,
                OPTION_CLIENT_IDENTIFIER = 61,
                OPTION_CLASSLESSROUTE = 121
                )

dhcp_message_type = enum('dhcp_message_type', globals(), uint8,
                        DHCPDISCOVER           = 1,
                        DHCPOFFER              = 2,
                        DHCPREQUEST            = 3,
                        DHCPDECLINE            = 4,
                        DHCPACK                = 5,
                        DHCPNAK                = 6,
                        DHCPRELEASE            = 7,
                        DHCPINFORM             = 8
                        )

dhcp_overload = enum('dhcp_overload', globals(), uint8, True,
                     OVERLOAD_FILE = 1,
                     OVERLOAD_SNAME = 2)

class _EmptyOptionException(Exception):
    pass

def _no_empty(l):
    if not l:
        raise _EmptyOptionException()
    return l

def _prepack_dhcp_option(x):
    if x.tag != OPTION_PAD and x.tag != OPTION_END:
        x.length = 0
        x.length = x._realsize() - 2
        if x.length > 255:
            x.length = 255
    else:
        if hasattr(x, 'length'):
            delattr(x, 'length')

dhcp_option = nstruct((dhcp_tag, 'tag'),
                      (optional(uint8, 'length', lambda x: x.tag != OPTION_PAD and x.tag != OPTION_END),),
                      name = 'dhcp_option',
                      size = lambda x: getattr(x, 'length', -1) + 2,
                      prepack = _prepack_dhcp_option,
                      padding = 1
                      )

dhcp_option_single_byte = nstruct(name = 'dhcp_option_single_byte',
                                  base = dhcp_option,
                                  criteria = lambda x: x.tag in (OPTION_PAD, OPTION_END),
                                  init = packvalue(OPTION_PAD, 'tag'))

dhcp_option_address = nstruct((ip4_addr, 'value'),
                                 name = 'dhcp_option_address',
                                 base = dhcp_option,
                                 criteria = lambda x: x.tag in (OPTION_NETMASK, OPTION_REQUESTED_IP, OPTION_SERVER_IDENTIFIER, OPTION_BROADCAST),
                                 init = packvalue(OPTION_NETMASK, 'tag'))

dhcp_option_address._parse_from_value = lambda x: ip4_addr(x)

dhcp_time_value = enum('dhcp_time_value', globals(), uint32,
                       DHCPTIME_INFINITE = 0xffffffff
                  )

dhcp_option_time = nstruct((dhcp_time_value, 'value'),
                             name = 'dhcp_option_time',
                             base = dhcp_option,
                             criteria = lambda x: x.tag in (OPTION_TIME_OFFSET, OPTION_LEASE_TIME, OPTION_T1, OPTION_T2),
                             init = packvalue(OPTION_LEASE_TIME, 'tag')
                             )

dhcp_option_time._parse_from_value = lambda x: int(x) if x is not None and x != 'infinite' else DHCPTIME_INFINITE

dhcp_option_servers = nstruct((ip4_addr[0], 'value'),
                              name = 'dhcp_option_servers',
                              base = dhcp_option,
                              criteria = lambda x: OPTION_ROUTER <= x.tag <= OPTION_RLSERVER or x.tag in (OPTION_NTPSERVER,),
                              init = packvalue(OPTION_ROUTER, 'tag')
                              )

try:
    unicode
except Exception:
    unicode = str

dhcp_option_servers._parse_from_value = lambda x: _no_empty([ip4_addr(x)] if isinstance(x, (str, unicode)) else [ip4_addr(v) for v in x])

dhcp_option_data = nstruct((raw, 'value'),
                           name = 'dhcp_option_name',
                           base = dhcp_option,
                           criteria = lambda x: x.tag in (OPTION_HOSTNAME, OPTION_DOMAINNAME, OPTION_MESSAGE, OPTION_CLIENT_IDENTIFIER),
                           init = packvalue(OPTION_HOSTNAME, 'tag'))


def _tobytes(s):
    if isinstance(s, bytes):
        return s
    elif isinstance(s, str):
        return s.decode('ascii')
    elif hasattr(s, '__iter__'):
        return create_binary(s, len(s))
    else:
        raise ValueError('Unsupported type: ')


dhcp_option_data._parse_from_value = _tobytes

dhcp_option_message_type = nstruct((dhcp_message_type, 'value'),
                                   name = 'dhcp_option_message_type',
                                   base = dhcp_option,
                                   criteria = lambda x: x.tag in (OPTION_MESSAGE_TYPE,),
                                   init = packvalue(OPTION_MESSAGE_TYPE, 'tag'))

dhcp_option_uint16 = nstruct((uint16, 'value'),
                          name = 'dhcp_option_uint16',
                          base = dhcp_option,
                          criteria = lambda x: x.tag in (OPTION_MTU, OPTION_MAX_MESSAGE_SIZE),
                          init = packvalue(OPTION_MTU, 'tag'))

dhcp_option_uint16._parse_from_value = lambda x: int(x)

dhcp_option_requested_options = nstruct((dhcp_tag[0], 'value'),
                                        name = 'dhcp_option_requested_options',
                                        base = dhcp_option,
                                        criteria = lambda x: x.tag == OPTION_REQUESTED_OPTIONS,
                                        init = packvalue(OPTION_REQUESTED_OPTIONS, 'tag'))


# option overload indicates that this DHCP packet use sname or file fields to send extra options
dhcp_option_overload = nstruct((dhcp_overload, 'value'),
                               name = 'dhcp_option_overload',
                               base = dhcp_option,
                               criteria = lambda x: x.tag == OPTION_OVERLOAD,
                               init = packvalue(OPTION_OVERLOAD, 'tag'))

def _prepack_destination(x):
    x.subnet = x.subnet[:(x.mask + 7) // 8]

def _destination_formatter(x):
    x['_cidr'] = ip4_addr_bytes.formatter(x.subnet) + '/' + str(x.mask)
    return x

_route_destination = nstruct((uint8, 'mask'),
                            (raw, 'subnet'),
                            name = '_route_destination',
                            padding = 1,
                            size = lambda x: (x.mask + 7) // 8 + 1,
                            prepack = _prepack_destination,
                            formatter = _destination_formatter
                            )

dhcp_route = nstruct((_route_destination,),
                     (ip4_addr, 'router'),
                     name = 'dhcp_route',
                     padding = 1
                     )

dhcp_classless_routes = nstruct((dhcp_route[0], 'value'),
                                name = 'dhcp_classless_routes',
                                base = dhcp_option,
                                criteria = lambda x: x.tag == OPTION_CLASSLESSROUTE,
                                init = packvalue(OPTION_CLASSLESSROUTE, 'tag'))

def _create_dhcp_route(cidr, router):
    network, mask = parse_ip4_network(cidr)
    return dhcp_route(subnet = ip4_addr.tobytes(network), mask = mask, router = ip4_addr(router))

dhcp_classless_routes._parse_from_value = lambda x: _no_empty([_create_dhcp_route(cidr, router) for cidr, router in x])

# According to RFC, options may be split into multiple parts
dhcp_option_partial = nstruct((dhcp_tag, 'tag'),
                      (optional(uint8, 'length', lambda x: x.tag != OPTION_PAD and x.tag != OPTION_END),),
                      (raw, 'data'),
                      name = 'dhcp_option_partial',
                      size = lambda x: getattr(x, 'length', -1) + 2,
                      prepack = _prepack_dhcp_option,
                      padding = 1
                      )

dhcp_flags = enum('dhcp_flags', globals(), uint16, True,
                  DHCPFLAG_BROADCAST = 1)

dhcp_payload = nstruct((boot_op, 'op'),
                       (uint8, 'htype'),
                       (uint8, 'hlen'),
                       (uint8, 'hops'),
                       (uint32, 'xid'),
                       (uint16, 'secs'),
                       (dhcp_flags, 'flags'),
                       (ip4_addr, 'ciaddr'),
                       (ip4_addr, 'yiaddr'),
                       (ip4_addr, 'siaddr'),
                       (ip4_addr, 'giaddr'),
                       (char[16], 'chaddr'),
                       (char[64], 'sname'),
                       (char[128], 'file'),
                       (uint32, 'magic_cookie'),    # 0x63825363
                       (dhcp_option_partial[0], 'options'),
                       name = 'dhcp_payload',
                       padding = 1,
                       extend = {'chaddr': mac_addr_bytes}
                       )

def reassemble_options(payload):
    '''
    Reassemble partial options to options, returns a list of dhcp_option
    
    DHCP options are basically `|tag|length|value|` structure. When an
    option is longer than 255 bytes, it can be splitted into multiple
    structures with the same tag. The splitted structures must be
    joined back to get the original option.
    
    `dhcp_option_partial` is used to present the splitted options,
    and `dhcp_option` is used for reassembled option.
    '''
    options = []
    option_indices = {}
    def process_option_list(partials):
        for p in partials:
            if p.tag == OPTION_END:
                break
            if p.tag == OPTION_PAD:
                continue
            if p.tag in option_indices:
                # Reassemble the data
                options[option_indices[p.tag]][1].append(p.data)
            else:
                options.append((p.tag, [p.data]))
                option_indices[p.tag] = len(options) - 1
    # First process options field
    process_option_list(payload.options)
    if OPTION_OVERLOAD in option_indices:
        # There is an overload option
        data = b''.join(options[option_indices[OPTION_OVERLOAD]][1])
        overload_option = dhcp_overload.create(data)
        if overload_option & OVERLOAD_FILE:
            process_option_list(dhcp_option_partial[0].create(payload.file))
        if overload_option & OVERLOAD_SNAME:
            process_option_list(dhcp_option_partial[0].create(payload.sname))
    def _create_dhcp_option(tag, data):
        opt = dhcp_option(tag = tag)
        opt._setextra(data)
        opt._autosubclass()
        return opt
    return [_create_dhcp_option(tag, b''.join(data)) for tag,data in options]

def build_options(payload, options, maxsize = 576, overload = OVERLOAD_FILE | OVERLOAD_SNAME, allowpartial = True):
    '''
    Split a list of options
    
    This is the reverse operation of `reassemble_options`, it splits `dhcp_option` into
    `dhcp_option_partial` if necessary, and set overload option if field overloading is
    used.
    
    :param options: a list of `dhcp_option`
    
    :param maxsize: Limit the maximum DHCP message size. If options cannot fit into the DHCP
                    message, specified fields are overloaded for options. If options cannot
                    fit after overloading, extra options are DROPPED if allowpartial = True.
                    
                    It is important to sort the dhcp options by priority.
    
    :param overload: fields that are allowed to be overloaded
    
    :param allowpartial: When options cannot fit into the DHCP message, allow the rest options
                         to be dropped.
                         
    :return: Number of options that are dropped i.e. `options[:-return_value]` are dropped
    '''
    if maxsize < 576:
        maxsize = 576
    max_options_size = maxsize - 240
    # Ignore OPTION_PAD and OPTION_END
    options = [o for o in options if o.tag not in (OPTION_PAD, OPTION_END)]
    # Only preserve data
    option_data = [(o.tag, o._tobytes()[2:]) for o in options]
    def split_options(option_data, limits):
        """
        Split options into multiple fields
        
        :param option_data: list of (tag, data) pair
        
        :param limits: list of int for limit of each field (excluding PAD and END)
        
        :return: number of options that are dropped
        """
        # List of (dhcp_option_partial, option_not_finished)
        partial_options = []
        buffers = [0]
        if not options:
            return ([], 0)
        def create_result():
            # Remove any unfinished partial options
            while partial_options and partial_options[-1][1]:
                partial_options.pop()
            buffers.append(len(partial_options))
            r = [[po for po,_ in partial_options[buffers[i]:buffers[i+1]]] for i in range(0, len(buffers) - 1)]
            # Remove empty fields
            while r and not r[-1]:
                r.pop()
            return r
        # Current field used size
        current_size = 0
        limit_iter = iter(limits)
        try:
            next_limit = next(limit_iter)
        except (StopIteration, GeneratorExit):
            return ([], False)
        for i, (tag, data) in enumerate(option_data):
            # Current used data size
            data_size = 0
            # Do not split very small options on boundary, this may prevent some broken DHCP clients/servers
            # to cause problem
            nosplit = (len(data) <= 32)
            while True:
                # next partial option size should be:
                # 1. no more than the current field limit (minus 1-byte tag and 1-byte length)
                # 2. no more than the single dhcp_option_partial data limit (255 due to single byte length)
                # 3. no more than the rest data size
                next_size = min(next_limit - current_size - 2, 255, len(data) - data_size)
                if next_size < 0 or (next_size == 0 and data_size < len(data)) \
                        or (next_size < len(data) - data_size and nosplit):
                    # Cannot put this part of data on the current field, find the next field
                    try:
                        next_limit = next(limit_iter)
                    except (StopIteration, GeneratorExit):
                        return (create_result(), len(option_data) - i)
                    # Record field boundary
                    buffers.append(len(partial_options))
                    current_size = 0
                else:
                    # Put this partial option on current field
                    partial_options.append((dhcp_option_partial(tag = tag, data = data[data_size : data_size + next_size]),
                                            (next_size < len(data) - data_size)))
                    data_size += next_size
                    current_size += next_size + 2
                    if data_size >= len(data):
                        # finished current option
                        break
        return (create_result(), 0)
    # First try to fit all options in options field
    # preserve a byte for OPTION_END
    result, not_finished = split_options(option_data, [max_options_size - 1])
    if not_finished:
        if overload & (OVERLOAD_FILE | OVERLOAD_SNAME):
            # Try overload
            # minus a overload option (1-byte tag, 1-byte lenght, 1-byte dhcp_overload) and 1-byte OPTION_END
            limits = [max_options_size - 4]
            if overload & OVERLOAD_FILE:
                # preserve a byte for OPTION_END
                limits.append(127)
            if overload & OVERLOAD_SNAME:
                # preserve a byte for OPTION_END
                limits.append(63)
            result2, not_finished2 = split_options(option_data, limits)
            # Only overload if we have a better result
            if len(result2) > 1:
                result = result2
                not_finished = not_finished2
    if not allowpartial and not_finished:
        raise ValueError("%d options cannot fit into a DHCP message" % (not_finished,))
    if not result:
        return not_finished
    elif len(result) <= 1:
        # No overload
        payload.options = result[0] + [dhcp_option_partial(tag = OPTION_END)]
    else:
        overload_option = 0
        if len(result) >= 2 and result[1]:
            overload_option |= OVERLOAD_FILE
            # overload file field
            payload.file = dhcp_option_partial[0].tobytes(result[1] + [dhcp_option_partial(tag = OPTION_END)])
        if len(result) >= 3 and result[2]:
            overload_option |= OVERLOAD_SNAME
            # overload sname field
            payload.sname = dhcp_option_partial[0].tobytes(result[2] + [dhcp_option_partial(tag = OPTION_END)])
        # Put an overload option before any other options
        payload.options = [dhcp_option_partial(tag = OPTION_OVERLOAD, data = dhcp_overload.tobytes(overload_option))] \
                        + result[0] + [dhcp_option_partial(tag = OPTION_END)]
    return not_finished


def create_option_from_value(tag, value):
    """
    Set DHCP option with human friendly value
    """
    dhcp_option.parser()
    fake_opt = dhcp_option(tag = tag)
    for c in dhcp_option.subclasses:
        if c.criteria(fake_opt):
            if hasattr(c, '_parse_from_value'):
                return c(tag = tag, value = c._parse_from_value(value))
            else:
                raise ValueError('Invalid DHCP option ' + str(tag) + ": " + repr(value))
    else:
        fake_opt._setextra(_tobytes(value))
        return fake_opt


def create_dhcp_options(input_dict, ignoreError = False, generateNone = False):
    """
    Try best to create dhcp_options from human friendly values, ignoring
    invalid values
    """
    retdict = {}
    for k,v in dict(input_dict).items():
        try:
            if generateNone and v is None:
                retdict[k] = None
            else:
                try:
                    retdict[k] = create_option_from_value(k, v)
                except _EmptyOptionException:
                    if generateNone:
                        retdict[k] = None
        except Exception:
            if ignoreError:
                continue
            else:
                raise
    return retdict


dhcp_option_tag_name_type = extra()


def _convert_to_tag(tag_name):
    tag = dhcp_tag.getValue(tag_name)
    if tag is None:
        raise TypeMismatchException(tag_name, dhcp_option_tag_name_type, "Not a valid DHCP option name")
    return tag


dhcp_option_tag_name_type.bind(str, convert=_convert_to_tag)

dhcp_option_tag_type = (int, dhcp_option_tag_name_type)


dhcp_options_type = extra()


def _check_dhcp_options(options):
    try:
        create_dhcp_options(options)
    except Exception as e:
        raise TypeMismatchException(options, dhcp_options_type, str(e))


dhcp_options_type.bind([tuple_((dhcp_option_tag_type, object))],
                        check=_check_dhcp_options)



'''
Created on 2016/2/23

:author: hubo
'''


from vlcp.config import defaultconfig
from vlcp.server.module import Module, api, depend, call_api, ModuleNotification
from vlcp.event.runnable import RoutineContainer
from vlcp.service.sdn import ofpmanager
import vlcp.protocol.openflow.defs.openflow13 as of13
from vlcp.event.connection import ConnectionResetException
from vlcp.protocol.openflow.openflow import OpenflowProtocolException,\
    OpenflowAsyncMessageEvent
from vlcp.event.event import Event, withIndices, M_
from contextlib import closing

def _bytes(s):
    return s.encode('ascii')

@withIndices('connection')
class OpenflowPortSynchronized(Event):
    pass

class OpenflowPortNotAppearException(Exception):
    pass

@defaultconfig
@depend(ofpmanager.OpenflowManager)
class OpenflowPortManager(Module):
    '''
    Manage Ports from Openflow Protocol
    '''
    service = True
    def __init__(self, server):
        Module.__init__(self, server)
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._manage_ports
        self.routines.append(self.apiroutine)
        self.managed_ports = {}
        self.createAPI(api(self.getports, self.apiroutine),
                       api(self.getallports, self.apiroutine),
                       api(self.getportbyno, self.apiroutine),
                       api(self.waitportbyno, self.apiroutine),
                       api(self.getportbyname, self.apiroutine),
                       api(self.waitportbyname, self.apiroutine),
                       api(self.resync, self.apiroutine)
                       )
        self._synchronized = False

    async def _get_ports(self, connection, protocol, onup = False, update = True):
        ofdef = connection.openflowdef
        dpid = connection.openflow_datapathid
        vhost = connection.protocol.vhost
        add = []
        try:
            if hasattr(ofdef, 'ofp_multipart_request'):
                # Openflow 1.3, use ofp_multipart_request to get ports
                openflow_reply = await protocol.querymultipart(ofdef.ofp_multipart_request(type=ofdef.OFPMP_PORT_DESC),
                                                               connection, self.apiroutine)
                ports = self.managed_ports.setdefault((vhost, dpid), {})
                for msg in openflow_reply:
                    for p in msg.ports:
                        add.append(p)
                        ports[p.port_no] = p
            else:
                # Openflow 1.0, use features_request
                if onup:
                    # Use the features_reply on connection setup
                    reply = connection.openflow_featuresreply
                else:
                    request = ofdef.ofp_msg()
                    request.header.type = ofdef.OFPT_FEATURES_REQUEST
                    reply = await protocol.querywithreply(request)
                ports = self.managed_ports.setdefault((vhost, dpid), {})
                for p in reply.ports:
                    add.append(p)
                    ports[p.port_no] = p
            if update:
                await self.apiroutine.wait_for_send(OpenflowPortSynchronized(connection))
                self._logger.info("Openflow port synchronized on connection %r", connection)
                await self.apiroutine.wait_for_send(ModuleNotification(self.getServiceName(), 'update',
                                                                         datapathid = connection.openflow_datapathid,
                                                                         connection = connection,
                                                                         vhost = protocol.vhost,
                                                                         add = add, remove = [],
                                                                         reason = 'connected'))
        except ConnectionResetException:
            pass
        except OpenflowProtocolException:
            pass

    async def _get_existing_ports(self):
        r =  await call_api(self.apiroutine, 'openflowmanager', 'getallconnections', {'vhost':None})
        await self.apiroutine.execute_all([self._get_ports(c, c.protocol, False, False)
                                           for c in r
                                           if c.openflow_auxiliaryid == 0])
        self._synchronized = True
        self._logger.info("Openflow ports synchronized")
        await self.apiroutine.wait_for_send(ModuleNotification(self.getServiceName(), 'synchronized'))

    async def _wait_for_sync(self):
        if not self._synchronized:
            await ModuleNotification.createMatcher(self.getServiceName(), 'synchronized')

    async def _manage_ports(self):
        try:
            self.apiroutine.subroutine(self._get_existing_ports())
            conn_update = ModuleNotification.createMatcher('openflowmanager', 'update')
            port_status = OpenflowAsyncMessageEvent.createMatcher(of13.OFPT_PORT_STATUS, None, 0)
            while True:
                e, m_ = await M_(conn_update, port_status)
                if m_ is port_status:
                    m = e.message
                    c = e.connection
                    if (c.protocol.vhost, c.openflow_datapathid) in self.managed_ports:
                        if m.reason == c.openflowdef.OFPPR_ADD:
                            # A new port is added
                            self.managed_ports[(c.protocol.vhost, c.openflow_datapathid)][m.desc.port_no] = m.desc
                            self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'update',
                                                                         datapathid = c.openflow_datapathid,
                                                                         connection = c,
                                                                         vhost = c.protocol.vhost,
                                                                         add = [m.desc], remove = [],
                                                                         reason = 'add'))
                        elif m.reason == c.openflowdef.OFPPR_DELETE:
                            try:
                                del self.managed_ports[(c.protocol.vhost, c.openflow_datapathid)][m.desc.port_no]
                                self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'update',
                                                                             datapathid = c.openflow_datapathid,
                                                                             connection = c,
                                                                             vhost = c.protocol.vhost,
                                                                             add = [], remove = [m.desc],
                                                                             reason = 'delete'))
                            except KeyError:
                                pass
                        elif m.reason == c.openflowdef.OFPPR_MODIFY:
                            try:
                                self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'modified',
                                                                             datapathid = c.openflow_datapathid,
                                                                             connection = c,
                                                                             vhost = c.protocol.vhost,
                                                                             old = self.managed_ports[(c.protocol.vhost, c.openflow_datapathid)][m.desc.port_no],
                                                                             new = m.desc,
                                                                             reason = 'modified'))
                            except KeyError:
                                self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'update',
                                                                             datapathid = c.openflow_datapathid,
                                                                             connection = c,
                                                                             vhost = c.protocol.vhost,
                                                                             add = [m.desc], remove = [],
                                                                             reason = 'add'))
                            self.managed_ports[(c.protocol.vhost, c.openflow_datapathid)][m.desc.port_no] = m.desc
                else:
                    for c in e.remove:
                        if c.openflow_auxiliaryid == 0 and (c.protocol.vhost, c.openflow_datapathid) in self.managed_ports:
                            self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'update',
                                                 datapathid = c.openflow_datapathid,
                                                 connection = c,
                                                 vhost = c.protocol.vhost,
                                                 add = [], remove = list(self.managed_ports[(c.protocol.vhost, c.openflow_datapathid)].values()),
                                                 reason = 'disconnected'))
                            del self.managed_ports[(c.protocol.vhost, c.openflow_datapathid)]
                    for c in e.add:
                        if c.openflow_auxiliaryid == 0:
                            self.apiroutine.subroutine(self._get_ports(c, c.protocol, True, True))
        finally:
            self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'unsynchronized'))

    async def getports(self, datapathid, vhost = ''):
        "Return all ports of a specifed datapath"
        await self._wait_for_sync()
        r = self.managed_ports.get((vhost, datapathid))
        if r is None:
            return None
        else:
            return list(r.values())

    async def getallports(self, vhost = None):
        "Return all ``(datapathid, port, vhost)`` tuples, optionally filterd by vhost"
        await self._wait_for_sync()
        if vhost is None:
            return [(dpid, p, vh) for (vh, dpid),v in self.managed_ports.items() for p in v.values()]
        else:
            return [(dpid, p, vh) for (vh, dpid),v in self.managed_ports.items() if vh == vhost for p in v.values()]

    async def getportbyno(self, datapathid, portno, vhost = ''):
        "Return port with specified OpenFlow portno"
        await self._wait_for_sync()
        return self._getportbyno(datapathid, portno, vhost)

    def _getportbyno(self, datapathid, portno, vhost = ''):
        ports = self.managed_ports.get((vhost, datapathid))
        if ports is None:
            return None
        else:
            return ports.get(portno)

    async def waitportbyno(self, datapathid, portno, timeout = 30, vhost = ''):
        """
        Wait for the specified OpenFlow portno to appear, or until timeout.
        """
        await self._wait_for_sync()
        async def waitinner():
            ports = self.managed_ports.get((vhost, datapathid))
            if ports is None:
                c = await call_api(self.apiroutine, 'openflowmanager', 'waitconnection',
                                   {'datapathid': datapathid, 'vhost':vhost, 'timeout': timeout})
                ports = self.managed_ports.get((vhost, datapathid))
                if ports is None:
                    await OpenflowPortSynchronized.createMatcher(c)
                ports = self.managed_ports.get((vhost, datapathid))
                if ports is None:
                    raise ConnectionResetException('Datapath %016x is not connected' % datapathid)
            if portno not in ports:
                ev = await OpenflowAsyncMessageEvent.createMatcher(of13.OFPT_PORT_STATUS, datapathid, 0,
                                                                   _ismatch = lambda x: x.message.desc.port_no == portno)
                return ev.message.desc
            else:
                return ports[portno]
        timeout_, result = await self.apiroutine.execute_with_timeout(timeout, waitinner())
        if timeout_:
            raise OpenflowPortNotAppearException('Port %d does not appear on datapath %016x' % (portno, datapathid))
        else:
            return result

    async def getportbyname(self, datapathid, name, vhost = ''):
        "Return port with specified port name"
        await self._wait_for_sync()
        return self._getportbyname(datapathid, name, vhost)

    def _getportbyname(self, datapathid, name, vhost = ''):
        if not isinstance(name, bytes):
            name = _bytes(name)
        ports = self.managed_ports.get((vhost, datapathid))
        if ports is None:
            return None
        else:
            for p in ports.values():
                if p.name == name:
                    return p
            return None

    async def waitportbyname(self, datapathid, name, timeout = 30, vhost = ''):
        """
        Wait for a port with the specified port name to appear, or until timeout
        """
        await self._wait_for_sync()
        if not isinstance(name, bytes):
            name = _bytes(name)
        async def waitinner():
            ports = self.managed_ports.get((vhost, datapathid))
            if ports is None:
                c = await call_api(self.apiroutine, 'openflowmanager', 'waitconnection',
                                   {'datapathid': datapathid, 'vhost':vhost, 'timeout': timeout})
                ports = self.managed_ports.get((vhost, datapathid))
                if ports is None:
                    await OpenflowPortSynchronized.createMatcher(c)
                ports = self.managed_ports.get((vhost, datapathid))
                if ports is None:
                    raise ConnectionResetException('Datapath %016x is not connected' % datapathid)
            for p in ports.values():
                if p.name == name:
                    return p
            ev = await OpenflowAsyncMessageEvent.createMatcher(of13.OFPT_PORT_STATUS, datapathid, 0,
                                                               _ismatch = lambda x: x.message.desc.name == name)
            return ev.message.desc
        timeout_, result = await self.apiroutine.execute_with_timeout(timeout, waitinner())
        if timeout_:
            raise OpenflowPortNotAppearException('Port %r does not appear on datapath %016x' % (name, datapathid))
        else:
            return result

    async def resync(self, datapathid, vhost = ''):
        '''
        Resync with current ports
        '''
        # Sometimes when the OpenFlow connection is very busy, PORT_STATUS message may be dropped.
        # We must deal with this and recover from it
        # Save current manged_ports
        if (vhost, datapathid) not in self.managed_ports:
            return
        else:
            last_ports = set(self.managed_ports[(vhost, datapathid)].keys())
        add = set()
        remove = set()
        ports = {}
        for _ in range(0, 10):
            c = await call_api(self.apiroutine, 'openflowmanager', 'getconnection',
                               {'datapathid': datapathid, 'vhost':vhost})
            if c is None:
                # Disconnected, will automatically resync when reconnected
                return
            ofdef = c.openflowdef
            protocol = c.protocol
            try:
                if hasattr(ofdef, 'ofp_multipart_request'):
                    # Openflow 1.3, use ofp_multipart_request to get ports
                    result = await protocol.querymultipart(ofdef.ofp_multipart_request(type=ofdef.OFPMP_PORT_DESC),
                                                           c, self.apiroutine)
                    for msg in result:
                        for p in msg.ports:
                            ports[p.port_no] = p
                else:
                    # Openflow 1.0, use features_request
                    request = ofdef.ofp_msg()
                    request.header.type = ofdef.OFPT_FEATURES_REQUEST
                    reply = await protocol.querywithreply(request)
                    for p in reply.ports:
                        ports[p.port_no] = p
            except ConnectionResetException:
                break
            except OpenflowProtocolException:
                break
            else:
                if (vhost, datapathid) not in self.managed_ports:
                    return
                current_ports = set(self.managed_ports[(vhost, datapathid)])
                # If a port is already removed
                remove.intersection_update(current_ports)
                # If a port is already added
                add.difference_update(current_ports)
                # If a port is not acquired, we do not add it
                acquired_keys = set(ports.keys())
                add.difference_update(acquired_keys)
                # Add and remove previous added/removed ports
                current_ports.difference_update(remove)
                current_ports.update(add)
                # If there are changed ports, the changed ports may or may not appear in the acquired port list
                # We only deal with following situations:
                # 1. If both lack ports, we add them
                # 2. If both have additional ports, we remote them
                to_add = acquired_keys.difference(current_ports.union(last_ports))
                to_remove = current_ports.intersection(last_ports).difference(acquired_keys)
                if not to_add and not to_remove and current_ports == last_ports:
                    break
                else:
                    add.update(to_add)
                    remove.update(to_remove)
                    current_ports.update(to_add)
                    current_ports.difference_update(to_remove)
                    last_ports = current_ports
        # Actual add and remove
        mports = self.managed_ports[(vhost, datapathid)]
        add_ports = []
        remove_ports = []
        for k in add:
            if k not in mports:
                add_ports.append(ports[k])
            mports[k] = ports[k]
        for k in remove:
            try:
                oldport = mports.pop(k)
            except KeyError:
                pass
            else:
                remove_ports.append(oldport)
        await self.apiroutine.wait_for_send(ModuleNotification(self.getServiceName(), 'update',
                                                                 datapathid = datapathid,
                                                                 connection = c,
                                                                 vhost = vhost,
                                                                 add = add_ports, remove = remove_ports,
                                                                 reason = 'resync'))



from vlcp.server.module import Module, publicapi
from vlcp.event.runnable import RoutineContainer
from vlcp.utils.networkmodel import PhysicalNetworkMap, PhysicalNetwork,\
    VXLANEndpointSet
from vlcp.config.config import defaultconfig
from vlcp.utils.networkplugin import createphysicalnetwork,\
    updatephysicalnetwork, default_physicalnetwork_keys, deletephysicalnetwork,\
    deletephysicalport, default_physicalport_keys, createlogicalnetwork,\
    default_logicalnetwork_keys, default_processor, updatelogicalnetwork,\
    deletelogicalnetwork, createphysicalport, updatephysicalport,\
    default_logicalnetwork_delete_check
from vlcp.utils.exceptions import WalkKeyNotRetrieved
from vlcp.utils.walkerlib import ensure_keys


@defaultconfig
class NetworkVxlanDriver(Module):
    """
    Network driver for VXLAN networks. When creating a VXLAN type physical network,
    you must specify an extra option ``vnirange``.
    """
    def __init__(self,server):
        super(NetworkVxlanDriver,self).__init__(server)
        self.app_routine = RoutineContainer(self.scheduler)
        self.app_routine.main = self._main
        self.routines.append(self.app_routine)
        self.createAPI(
                       publicapi(self.createphysicalnetwork,
                                    criteria=lambda type: type == 'vxlan'),
                       publicapi(self.updatephysicalnetwork,
                                    criteria=lambda type: type == 'vxlan'),
                       publicapi(self.deletephysicalnetwork,
                                    criteria=lambda type: type == 'vxlan'),
                       publicapi(self.createphysicalport,
                                    criteria=lambda type: type == 'vxlan'),
                       publicapi(self.updatephysicalport,
                                    criteria=lambda type: type == 'vxlan'),
                       publicapi(self.deletephysicalport,
                                    criteria=lambda type: type == 'vxlan'),
                       publicapi(self.createlogicalnetwork,
                                    criteria=lambda type: type == 'vxlan'),
                       publicapi(self.updatelogicalnetwork,
                                    criteria=lambda type: type == 'vxlan'),
                       publicapi(self.deletelogicalnetwork,
                                    criteria=lambda type: type == "vxlan")
                       )

    async def _main(self):
        self._logger.info("network_vxlan_driver running ---")

    def createphysicalnetwork(self, type):
        # create an new physical network
        def create_physicalnetwork_processor(physicalnetwork, walk, write, *, parameters):
            if 'vnirange' not in parameters:
                raise ValueError('must specify vnirange with physical network type=vxlan')
            _check_vnirange(parameters['vnirange'])
            return default_processor(physicalnetwork, parameters=parameters, excluding=('id', 'type'))
        return createphysicalnetwork(type, create_processor=create_physicalnetwork_processor),\
                default_physicalnetwork_keys

    def updatephysicalnetwork(self, type):
        # update a physical network
        def update_physicalnetwork_keys(id_, parameters):
            if 'vnirange' in parameters:
                return (PhysicalNetworkMap.default_key(id_),)
            else:
                return ()
        def update_physicalnetwork_processor(physicalnetwork, walk, write, *, parameters):
            if 'vnirange' in parameters:
                _check_vnirange(parameters['vnirange'])
                try:
                    phymap = walk(PhysicalNetworkMap.default_key(physicalnetwork.id))
                except WalkKeyNotRetrieved:
                    pass
                else:
                    _check_vnirange_allocation(parameters['vnirange'], phymap.network_allocation)
            return default_processor(physicalnetwork, parameters=parameters, disabled=('type',))
        return updatephysicalnetwork(update_processor=update_physicalnetwork_processor), update_physicalnetwork_keys

    def deletephysicalnetwork(self, type):
        return deletephysicalnetwork(), default_physicalnetwork_keys

    def createphysicalport(self, type):
        return createphysicalport(), default_physicalport_keys

    def updatephysicalport(self, type):
        return updatephysicalport(), None

    def deletephysicalport(self, type):
        return deletephysicalport(), default_physicalport_keys

    def createlogicalnetwork(self, type):
        def logicalnetwork_processor(logicalnetwork, logicalnetworkmap, physicalnetwork,
                                     physicalnetworkmap, walk, write, *, parameters):
            if 'vni' in parameters:
                # Allocate this vni
                vni = int(parameters['vni'])
                if _isavaliablevni(physicalnetwork.vnirange, physicalnetworkmap.network_allocation, vni):
                    physicalnetworkmap.network_allocation[str(vni)] = logicalnetwork.create_weakreference()
                    write(physicalnetworkmap.getkey(), physicalnetworkmap)
                else:
                    raise ValueError("Specified VNI " + str(vni) + " allocated or not in range")
            else:
                # Allocate a vni from range
                vni = _findavaliablevni(physicalnetwork.vnirange, physicalnetworkmap.network_allocation)
                if vni is None:
                    raise ValueError("no available VNI in physical network " + physicalnetwork.id)
                physicalnetworkmap.network_allocation[str(vni)] = logicalnetwork.create_weakreference()
                write(physicalnetworkmap.getkey(), physicalnetworkmap)
            logicalnetwork.vni = vni
            write(logicalnetwork.getkey(), logicalnetwork)
            # Create VXLANEndpointSet
            vxlanendpoint_set = VXLANEndpointSet.create_instance(logicalnetwork.id)
            write(vxlanendpoint_set.getkey(), vxlanendpoint_set)
            return default_processor(logicalnetwork, parameters=parameters, excluding=('id', 'physicalnetwork', 'vni'))
        # Process logical networks with specified IDs first
        return createlogicalnetwork(create_processor=logicalnetwork_processor,
                                    reorder_dict=lambda x: sorted(x.items(), key=lambda y: 'vni' in y[1], reverse=True)),\
               default_logicalnetwork_keys

    def updatelogicalnetwork(self, type):
        # When updating VLAN ids, Must first deallocate all VLAN ids, then allocate all
        # Chaining two walkers for this
        def update_logicalnetwork_keys(id_, parameters):
            if 'vni' in parameters:
                return (PhysicalNetwork.default_key(id_),
                        PhysicalNetworkMap.default_key(id_))
            else:
                return ()
        def deallocate_processor(logicalnetwork, walk, write, *, parameters):
            if 'vni' in parameters:
                try:
                    phymap = walk(PhysicalNetworkMap._network.leftkey(logicalnetwork.physicalnetwork))
                except WalkKeyNotRetrieved:
                    pass
                else:
                    del phymap.network_allocation[str(logicalnetwork.vni)]
                    write(phymap.getkey(), phymap)
            return False
        deallocate_walker = updatelogicalnetwork(update_processor=deallocate_processor)
        def allocate_processor(logicalnetwork, walk, write, *, parameters):
            if 'vni' in parameters:
                try:
                    phynet = walk(logicalnetwork.physicalnetwork.getkey())
                    phymap = walk(PhysicalNetworkMap._network.leftkey(logicalnetwork.physicalnetwork))
                except WalkKeyNotRetrieved:
                    ensure_keys(PhysicalNetworkMap._network.leftkey(logicalnetwork.physicalnetwork))
                else:
                    vni = int(parameters['vni'])
                    if _isavaliablevni(phynet.vnirange, phymap.network_allocation, vni):
                        phymap.network_allocation[str(vni)] = logicalnetwork.create_weakreference()
                        write(phymap.getkey(), phymap)
                    else:
                        raise ValueError("Specified VNI " + str(vni) + " allocated or not in range")
                    logicalnetwork.vni = vni
                    write(logicalnetwork.getkey(), logicalnetwork)
            return default_processor(logicalnetwork, parameters=parameters, excluding=('id', 'vni'),
                                                                            disabled=('physicalnetwork',))
        allocate_walker = updatelogicalnetwork(update_processor=allocate_processor)
        def walker(walk, write, timestamp, parameters_dict):
            deallocate_walker(walk, write, timestamp, parameters_dict)
            allocate_walker(walk, write, timestamp, parameters_dict)
        return walker, update_logicalnetwork_keys

    def deletelogicalnetwork(self, type):
        def check_processor(logicalnetwork, logicalnetworkmap,
                            physicalnetwork, physicalnetworkmap,
                            walk, write, *, parameters):
            default_logicalnetwork_delete_check(logicalnetwork, logicalnetworkmap,
                                                physicalnetwork, physicalnetworkmap,
                                                walk, write, parameters=parameters)
            del physicalnetworkmap.network_allocation[str(logicalnetwork.vni)]
            write(physicalnetworkmap.getkey(), physicalnetworkmap)
        return deletelogicalnetwork(check_processor=check_processor), default_logicalnetwork_keys


def _check_vnirange(vnirange):
    lastend = 0
    for start,end in vnirange:
        if start <= 0 or end > (1 << 24) - 1:
            raise ValueError('VNI out of range (1 - 4095)')
        if start > end or start <= lastend:
            raise ValueError('VNI sequences overlapped or disordered: [%r, %r]' % (start, end))
        lastend = end


def _check_vnirange_allocation(vnirange, allocation):
    allocated_ids = sorted(int(k) for k in allocation.keys())
    range_iter = iter(vnirange)
    current_range = None
    for id_ in allocated_ids:
        while current_range is None or current_range[1] < id_:
            try:
                current_range = next(range_iter)
            except StopIteration:
                raise ValueError("Allocated VNI " + str(id_) + " not in new VNI range")
        if current_range[0] > id_:
            raise ValueError("Allocated VNI " + str(id_) + " not in new VNI range")


def _findavaliablevni(vnirange,allocated):
    
    vni = None
    for vr in vnirange:
        find = False
        for v in range(vr[0],vr[1] + 1):
            if str(v) not in allocated:
                vni = v
                find = True
                break

        if find:
            break
    return vni


def _isavaliablevni(vnirange,allocated,vni):
    
    find = False
    for start,end in vnirange:
        if start <= int(vni) <= end:
            find = True
            break

    if find:
        if str(vni) not in allocated:
            find = True
        else:
            find = False
    else:
        find = False

    return find



'''
Created on 2015/12/25

:author: hubo
'''

from vlcp.protocol import Protocol
from vlcp.config import defaultconfig
from vlcp.event import Event, withIndices, ConnectionWriteEvent
import logging
import os
from vlcp.event.stream import Stream, StreamDataEvent
from vlcp.event.connection import Client
from vlcp.event.event import M_

@withIndices('state', 'connection', 'connmark', 'createby')
class RawConnectionStateEvent(Event):
    CONNECTION_UP = 'up'
    CONNECTION_DOWN = 'down'
    CONNECTION_NOTCONNECTED = 'notconnected'

def _copy(buffer):
    try:
        if isinstance(buffer, memoryview):
            return buffer.tobytes()
        else:
            return buffer[:]
    except Exception:
        return buffer[:]

@defaultconfig
class Raw(Protocol):
    '''
    Raw protocol, provide two streams for input and output
    '''
    _default_persist = False
    _default_defaultport = 0
    _default_createqueue = True
    # Enable/disable buffering for the output stream.
    # It is dangerous to use buffering in output stream because small amount of data might
    # stay in buffer and not be sent
    _default_buffering = False
    _default_writebufferlimit = 4096
    # Split very large data to chunks to balance the output streaming
    _default_splitsize = 1048576
    _logger = logging.getLogger(__name__ + '.JsonRPC')
    def __init__(self):
        '''
        Constructor
        '''
        Protocol.__init__(self)
    async def init(self, connection):
        await Protocol.init(self, connection)
        connection.createdqueues.append(connection.scheduler.queue.addSubQueue(\
                self.messagepriority, RawConnectionStateEvent.createMatcher(connection = connection), ('connstate', connection)))
        await self.reconnect_init(connection)
    async def _raw_writer(self, connection):
        try:
            while True:
                await connection.outputstream.prepareRead(connection)
                try:
                    data = connection.outputstream.readonce()
                except EOFError:
                    await connection.write(ConnectionWriteEvent(connection, connection.connmark, data=b'', EOF=True))
                    break
                except IOError:
                    await connection.reset()
                    break
                else:
                    await connection.write(ConnectionWriteEvent(connection, connection.connmark, data=data, EOF=False))
        finally:
            connection.outputstream.close(connection.scheduler)
    async def reconnect_init(self, connection):
        connection.inputstream = Stream()
        connection.outputstream = Stream(writebufferlimit=(self.writebufferlimit if self.buffering else 0),
                                         splitsize=self.splitsize)
        connection.subroutine(self._raw_writer(connection), False, '_raw_writer')
        await connection.wait_for_send(RawConnectionStateEvent(RawConnectionStateEvent.CONNECTION_UP, connection, connection.connmark, self))
    async def notconnected(self, connection):
        await Protocol.notconnected(self, connection)
        await connection.waitForSend(RawConnectionStateEvent(RawConnectionStateEvent.CONNECTION_NOTCONNECTED, connection, connection.connmark, self))
    async def closed(self, connection):
        await Protocol.closed(self, connection)
        await connection.inputstream.write(b'', connection, True, True)
        connection.terminate(connection._raw_writer)
        connection._raw_writer = None
        await connection.wait_for_send(RawConnectionStateEvent(RawConnectionStateEvent.CONNECTION_DOWN, connection, connection.connmark, self))
    async def error(self, connection):
        await Protocol.error(self, connection)
        await connection.inputstream.error(connection, True)
        connection.terminate(connection._raw_writer)
        connection._raw_writer = None
        await connection.wait_for_send(RawConnectionStateEvent(RawConnectionStateEvent.CONNECTION_DOWN, connection, connection.connmark, self))
    def statematcher(self, connection, state = RawConnectionStateEvent.CONNECTION_DOWN, currentconn = True):
        if currentconn:
            return RawConnectionStateEvent.createMatcher(state, connection, connection.connmark)
        else:
            return RawConnectionStateEvent.createMatcher(state, connection)
    async def client_connect(self, container, url, *args, **kwargs):
        '''
        Create a connection with raw protocol
        
        :param container: current routine container
        
        :param url: url to connect to (see Client)
        
        :param \*args, \*\*kwargs: other parameters to create a Client (except url, protocol and scheduler)
        
        :returns: `(connection, inputstream, outputstream)` where client is the created connection, inputstream
                  is the stream to read from the socket, outputstream is the stream to write to socket
        '''
        c = Client(url, self, container.scheduler, *args, **kwargs)
        c.start()
        ev, m = await M_(self.statematcher(c, RawConnectionStateEvent.CONNECTION_UP, False), self.statematcher(c, RawConnectionStateEvent.CONNECTION_NOTCONNECTED, False))
        if ev.state == RawConnectionStateEvent.CONNECTION_UP:
            return (c, ev.inputstream, ev.outputstream)
        else:
            raise IOError('Connection failed')
    def redirect_outputstream(self, connection, stream):
        "Close current outputstream and output from the new stream"
        if not connection.connected:
            raise IOError('Connection is closed')
        if connection._raw_writer:
            connection.terminate(connection._raw_writer)
            del connection._raw_writer
        connection.outputstream = stream
        connection.subroutine(self._raw_writer(connection), False, '_raw_writer')
    def parse(self, connection, data, laststart):
        if connection.inputstream and not connection.inputstream.closed:
            if len(data) == laststart:
                return ([StreamDataEvent(connection.inputstream, StreamDataEvent.STREAM_EOF, data=b'')], 0)
            else:
                return ([StreamDataEvent(connection.inputstream, StreamDataEvent.STREAM_DATA, data=_copy(data))], 0)
        else:
            return ([], 0)



'''
Created on 2018/8/9

:author: wangqianb
'''

from vlcp.config.config import defaultconfig
from vlcp.server.module import depend, api
from vlcp.service.sdn.flowbase import FlowBase
from vlcp.service.sdn.ofpmanager import FlowInitialize
from vlcp.utils.flowupdater import FlowUpdater
from vlcp.utils.networkmodel import PhysicalPort, LogicalPort, LogicalNetwork,\
    LogicalNetworkMap
import vlcp.service.kvdb.objectdb as objectdb
import vlcp.service.sdn.ofpportmanager as ofpportmanager
from vlcp.event.runnable import RoutineContainer
from vlcp.protocol.openflow.openflow import OpenflowConnectionStateEvent,\
    OpenflowAsyncMessageEvent, OpenflowErrorResultException
from vlcp.utils.ethernet import arp_packet_l4, mac_addr, ethernet_l2, ip4_addr
import vlcp.service.sdn.ioprocessing as iop
import itertools
from vlcp.utils.exceptions import WalkKeyNotRetrieved
from contextlib import suppress
from vlcp.event.event import M_
from vlcp.server.module import call_api


class FreeARPUpdater(FlowUpdater):
    def __init__(self, connection, parent):
        FlowUpdater.__init__(self, connection, (), ('FreeARPUpdater', connection), parent._logger)
        self._parent = parent
        self._lastlognets = ()
        self._lastphyports = ()
        self._lastlogports = ()
        self._lastlogportinfo = {}
        self._lastphyportinfo = {}
        self._lastlognetinfo = {}
        self._last_arps = set()

    async def main(self):
        try:
            if self._connection.protocol.disablenxext:
                self._logger.warning("Free arp fuc disabled on connection %r because Nicira extension is not enabled", self._connection)
                return
            self.subroutine(self._update_handler(), True, '_update_handler_routine')
            await FlowUpdater.main(self)
        finally:
            if hasattr(self, '_update_handler_routine'):
                self._update_handler_routine.close()

    async def _update_handler(self):
        dataobjectchanged = iop.DataObjectChanged.createMatcher(None, None, self._connection)
        while True:
            ev = await dataobjectchanged
            self._lastlogports, self._lastphyports, self._lastlognets, _ = ev.current
            self._update_walk()
            self.updateobjects((p for p, _ in self._lastlogports))

    def _walk_logport(self, key, value, walk, save):
        save(key)

    def _walk_phyport(self, key, value, walk, save):
        save(key)

    def _walk_lognet(self, key, value, walk, save):
        save(key)


    def _update_walk(self):
        logport_keys = [p.getkey() for p, _ in self._lastlogports]
        phyport_keys = [p.getkey() for p, _ in self._lastphyports]
        lognet_keys = [n.getkey() for n, _ in self._lastlognets]
        lognet_mapkeys = [LogicalNetworkMap.default_key(n.id) for n, _ in self._lastlognets]

        self._initialkeys = logport_keys + lognet_keys + phyport_keys
        self._walkerdict = dict(itertools.chain(((n, self._walk_lognet) for n in lognet_keys),
                                                ((p, self._walk_logport) for p in logport_keys),
                                                ((p, self._walk_phyport) for p in phyport_keys)))
        self.subroutine(self.restart_walk(), False)

    async def updateflow(self, connection, addvalues, removevalues, updatedvalues):
        try:
            allobjs = set(o for o in self._savedresult if o is not None and not o.isdeleted())
            lastlogportinfo = self._lastlogportinfo
            currentlognetinfo = dict((n, (id, n.physicalnetwork)) for n, id in self._lastlognets if n in allobjs)
            currentlogportinfo = dict((p, (id, p.network)) for p, id in self._lastlogports if p in allobjs)
            currentphyportinfo = dict((p, (id, p.physicalnetwork)) for p, id in self._lastphyports if p in allobjs)

            currentphynetinfo = dict((p.physicalnetwork, (id, p)) for p, id in self._lastphyports if p in allobjs)
            self._lastlogportinfo = currentlogportinfo
            self._lastlognetinfo = currentlognetinfo
            self._lastphyportinfo = currentphyportinfo
            cmds = []
            ofdef = connection.openflowdef

            def packet_out_free_arp(logicalport,logicalnet,logicalnetid,phyportid):
                egress = self._parent._gettableindex("egress", self._connection.protocol.vhost)
                arp_request_packet = arp_packet_l4(
                    dl_src=mac_addr(logicalport.mac_address),
                    dl_dst=mac_addr("FF:FF:FF:FF:FF:FF"),
                    arp_op=ofdef.ARPOP_REQUEST,
                    arp_sha=mac_addr(logicalport.mac_address),
                    arp_spa=ip4_addr(logicalport.ip_address),
                    arp_tpa=ip4_addr(logicalport.ip_address)
                )

                return ofdef.ofp_packet_out(buffer_id=ofdef.OFP_NO_BUFFER,
                                                 in_port=ofdef.OFPP_CONTROLLER,
                                                 actions=[
                                                     ofdef.ofp_action_set_field(
                                                         field=ofdef.create_oxm(ofdef.NXM_NX_REG4,
                                                                                logicalnetid)
                                                     ),
                                                     ofdef.ofp_action_set_field(
                                                         field=ofdef.create_oxm(ofdef.NXM_NX_REG5,
                                                                                logicalnetid)
                                                     ),
                                                     ofdef.ofp_action_set_field(
                                                         field=ofdef.create_oxm(ofdef.NXM_NX_REG6,
                                                                                phyportid)
                                                     ),
                                                     ofdef.ofp_action_set_field(
                                                         field=ofdef.create_oxm(ofdef.NXM_NX_REG7,
                                                                                0x4000)
                                                     ),
                                                     ofdef.nx_action_resubmit(
                                                         table=egress
                                                     )
                                                 ],
                                                 data=arp_request_packet._tobytes()
                                                 )

            for p in currentlogportinfo:
                if p not in lastlogportinfo or lastlogportinfo[p] != currentlogportinfo[p]:
                    _, logicalnet = currentlogportinfo[p]
                    logicalnetid, _ = currentlognetinfo[logicalnet]
                    phyportid, _ = currentphynetinfo[logicalnet.physicalnetwork]
                    if phyportid is not None and logicalnet.physicalnetwork.type in self._parent.allowednetworktypes:
                        if await call_api(self, 'ioprocessing', 'flowready', {'connection': connection,
                                                                            'logicalnetworkid': logicalnetid,
                                                                            'physicalportid': phyportid}):
                            cmds.append(packet_out_free_arp(p,logicalnet,logicalnetid,phyportid))
            if cmds is not None:
                await self.execute_commands(connection, cmds)
        except Exception:
            self._logger.warning("Unexpected exception in FreeARPUpdater. Will ignore and continue.", exc_info=True)


@defaultconfig
@depend(ofpportmanager.OpenflowPortManager, objectdb.ObjectDB)
class FreeARP(FlowBase):
    "Send FREE ARP"

    _tablerequest = (("ingress", (), ''),
                     ("egress", ("ingress",), ''))

    _default_allowednetworktypes = ("native", "vlan")

    def __init__(self, server):
        FlowBase.__init__(self, server)
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._main
        self.routines.append(self.apiroutine)
        self._flowupdaters = {}

    async def _main(self):
        flow_init = FlowInitialize.createMatcher(_ismatch = lambda x: self.vhostbind is None or x.vhost in self.vhostbind)
        conn_down = OpenflowConnectionStateEvent.createMatcher(state = OpenflowConnectionStateEvent.CONNECTION_DOWN,
                                                               _ismatch = lambda x: self.vhostbind is None or x.createby.vhost in self.vhostbind)
        while True:
            ev, m = await M_(flow_init, conn_down)
            if m is flow_init:
                c = ev.connection
                self.apiroutine.subroutine(self._init_conn(c))
            else:
                c = ev.connection
                self.apiroutine.subroutine(self._remove_conn(c))

    async def _init_conn(self, conn):
        # Default
        if conn in self._flowupdaters:
            updater = self._flowupdaters.pop(conn)
            updater.close()
        updater = FreeARPUpdater(conn, self)
        self._flowupdaters[conn] = updater
        updater.start()

    async def _remove_conn(self, conn):
        # Do not need to modify flows
        if conn in self._flowupdaters:
            self._flowupdaters.pop(conn).close()


'''
Created on 2016/2/19

:author: hubo
'''

from vlcp.config import defaultconfig
from vlcp.protocol.jsonrpc import JsonRPC, JsonRPCRequestEvent,\
    JsonRPCNotificationEvent
from vlcp.event.connection import ConnectionResetException, ConnectionWriteEvent
from contextlib import closing
from vlcp.event.event import M_

@defaultconfig
class OVSDB(JsonRPC):
    '''
    OVSDB protocol, this is a specialized JSON-RPC 1.0 protocol
    '''
    _default_defaultport = 6632
    # Only accept "echo" requests
    _default_allowedrequests = ('echo',)
    # Send "echo" requests when connection is idle
    _default_keepalivetime = 10
    # Disconnect when reply is not received for "echo" requests
    _default_keepalivetimeout = 3
    _default_tcp_nodelay = True
    async def _respond_echo(self, connection):
        try:
            request_matcher = JsonRPCRequestEvent.createMatcher('echo', connection)
            connstate = self.statematcher(connection)
            while True:
                ev, m = await M_(request_matcher, connstate)
                if m is request_matcher:
                    ev.canignore = True
                    reply = self.formatreply(ev.params, ev.id, connection)
                    reply.echoreply = True
                    await connection.write(reply, False)
                else:
                    break
        except ConnectionResetException:
            pass
        
    async def _extra_queues(self, connection):
        connection.createdqueues.append(connection.scheduler.queue.addSubQueue(\
                self.writepriority + 10, ConnectionWriteEvent.createMatcher(connection = connection, _ismatch = lambda x: hasattr(x, 'echoreply') and x.echoreply), ('echoreply', connection)))
        
    async def reconnect_init(self, connection):
        await JsonRPC.reconnect_init(self, connection)
        connection.subroutine(self._respond_echo(connection))

    async def keepalive(self, connection):
        try:
            timeout, _ = await connection.execute_with_timeout(self.keepalivetimeout,
                                                               self.querywithreply('echo', [], connection, connection))
            if timeout:
                await connection.reset(True)
        except Exception:
            await connection.reset(True)



'''
Created on 2015/6/16

:author: hubo
'''
from __future__ import print_function, absolute_import, division
import sys
from .core import QuitException, TimerEvent, SystemControlEvent
from .event import Event, withIndices, M_, Diff_
import asyncio


class EventHandler(object):
    '''
    Runnable with an event handler model. 
    '''
    def __init__(self, scheduler = None, daemon = False):
        self.handlers = dict()
        self.scheduler = scheduler
        self.daemon = daemon
        self.registered = False
    def bind(self, scheduler):
        self.scheduler = scheduler
    def __iter__(self):
        '''
        Keep it like a iterator. Not very useful.
        '''
        return self
    def next(self):
        '''
        Keep it like a iterator. Not very useful.
        '''
        self.send(None)
    def __next__(self):
        '''
        Python 3 next
        '''
        self.send(None)
    def send(self, etup):
        '''
        Handle events
        '''
        return self.handlers[etup[1]](etup[0], self.scheduler)
    def _setDaemon(self):
        if not self.registered:
            self.registered = True
            self.scheduler.setDaemon(self, self.daemon)
    def registerHandler(self, matcher, handler):
        '''
        Register self to scheduler
        '''
        self.handlers[matcher] = handler
        self.scheduler.register((matcher,), self)
        self._setDaemon()
    def unregisterHandler(self, matcher):
        self.scheduler.unregister((matcher,), self)
        del self.handlers[matcher]
    def unregisterAllHandlers(self):
        self.scheduler.unregister(tuple(self.handlers.keys()), self)
        self.handlers.clear()
    def registerAllHandlers(self, handlerDict):
        '''
        Register self to scheduler
        '''
        self.handlers.update(handlerDict)
        if hasattr(handlerDict, 'keys'):
            self.scheduler.register(handlerDict.keys(), self)
        else:
            self.scheduler.register(tuple(h[0] for h in handlerDict), self)
        self._setDaemon()
    def close(self):
        self.scheduler.unregisterall(self)
        self.registered = False
    def registerExceptionHandler(self, handler):
        self.exceptionHandler = handler
    def registerQuitHandler(self, handler):
        self.quitHandler = handler
    def throw(self, typ, val = None, tb = None):
        if val is None:
            if isinstance(typ, type):
                val = typ()
            else:
                val = typ
                typ = type(val)
        if isinstance(val, QuitException):
            self.quitHandler(self.scheduler)
        else:
            self.exceptionHandler(val, self.scheduler)
    def exceptionHandler(self, val, scheduler):
        raise val
    def quitHandler(self, scheduler):
        raise StopIteration


class GeneratorExit_(BaseException):
    """
    Bypass PyPy3 bug
    """
    pass


def _close_generator(g):
    """
    PyPy 3 generator has a bug that calling `close` caused
    memory leak. Before it is fixed, use `throw` instead
    """
    if isinstance(g, generatorwrapper):
        g.close()
    elif _get_frame(g) is not None:
        try:
            g.throw(GeneratorExit_)
        except (StopIteration, GeneratorExit_):
            return
        else:
            raise RuntimeError("coroutine ignored GeneratorExit")


def _await(coroutine):
    """
    Return a generator
    """
    if hasattr(coroutine, '__await__'):
        return coroutine.__await__()
    else:
        return coroutine


@withIndices('type', 'routine')
class RoutineControlEvent(Event):
    DELEGATE_FINISHED = 'delegatefinished'

class IllegalMatchersException(Exception):
    pass


def _get_frame(obj):
    if hasattr(obj, 'cr_frame'):
        return obj.cr_frame
    else:
        return obj.gi_frame


class generatorwrapper(object):
    '''
    Default __repr__ of a generator is not readable, use a wrapper to improve the readability
    '''
    __slots__ = ('run', 'name', 'classname', '_iter')
    def __init__(self, run, name = 'coroutine', classname = 'routine'):
        self.run = run
        self.name = name
        self.classname = classname
        if hasattr(run, '__await__'):
            self._iter = run.__await__()
        else:
            self._iter = run
    def __iter__(self):
        return self._iter
    __await__ = __iter__
    def next(self):
        try:
            return next(self._iter)
        except StopIteration:
            raise StopIteration
    def __next__(self):
        try:
            return next(self._iter)
        except StopIteration:
            raise StopIteration
    def send(self, arg):
        try:
            return self._iter.send(arg)
        except StopIteration:
            raise StopIteration
    def throw(self, typ, val = None, tb = None):
        try:
            return self._iter.throw(typ, val, tb)
        except StopIteration:
            raise StopIteration
    def __repr__(self, *args, **kwargs):
        try:
            iterator = _get_frame(self.run).f_locals[self.name]
            try:
                return '<%s %r of %r at 0x%016X>' % (self.classname, iterator,
                                                       _get_frame(iterator).f_locals['self'],
                                                       id(iterator))
            except Exception:
                return '<%s %r at 0x%016X>' % (self.classname, iterator, id(iterator))
        except Exception:
            return repr(self.run)
    def close(self):
        return _close_generator(self.run)


def Routine(coroutine, scheduler, asyncStart = True, container = None, manualStart = False, daemon = False):
    """
    This wraps a normal coroutine to become a VLCP routine. Usually you do not need to call this yourself;
    `container.start` and `container.subroutine` calls this automatically.
    """
    def run():
        iterator = _await(coroutine)
        iterself = yield
        if manualStart:
            yield
        try:
            if asyncStart:
                scheduler.yield_(iterself)
                yield
            if container is not None:
                container.currentroutine = iterself
            if daemon:
                scheduler.setDaemon(iterself, True)
            try:
                matchers = next(iterator)
            except StopIteration:
                return
            while matchers is None:
                scheduler.yield_(iterself)
                yield
                try:
                    matchers = next(iterator)
                except StopIteration:
                    return
            try:
                scheduler.register(matchers, iterself)
            except Exception:
                try:
                    iterator.throw(IllegalMatchersException(matchers))
                except StopIteration:
                    pass
                raise
            while True:
                try:
                    etup = yield
                except GeneratorExit_:
                    raise
                except:
                    #scheduler.unregister(matchers, iterself)
                    lmatchers = matchers
                    t,v,tr = sys.exc_info()  # @UnusedVariable
                    if container is not None:
                        container.currentroutine = iterself
                    try:
                        matchers = iterator.throw(t,v)
                    except StopIteration:
                        return
                else:
                    #scheduler.unregister(matchers, iterself)
                    lmatchers = matchers
                    if container is not None:
                        container.currentroutine = iterself
                    try:
                        matchers = iterator.send(etup)
                    except StopIteration:
                        return
                while matchers is None:
                    scheduler.yield_(iterself)
                    yield
                    try:
                        matchers = next(iterator)
                    except StopIteration:
                        return
                try:
                    if hasattr(matchers, 'two_way_difference'):
                        reg, unreg = matchers.two_way_difference(lmatchers)
                    else:
                        reg = set(matchers).difference(lmatchers)
                        unreg = set(lmatchers).difference(matchers)
                    scheduler.register(reg, iterself)
                    scheduler.unregister(unreg, iterself)
                except Exception:
                    try:
                        iterator.throw(IllegalMatchersException(matchers))
                    except StopIteration:
                        pass
                    raise
        finally:
            # iterator.close() can be called in other routines, we should restore the currentroutine variable
            if container is not None:
                lastcurrentroutine = getattr(container, 'currentroutine', None)
                container.currentroutine = iterself
            else:
                lastcurrentroutine = None
            _close_generator(coroutine)
            if container is not None:
                container.currentroutine = lastcurrentroutine
            scheduler.unregisterall(iterself)
    r = generatorwrapper(run())
    next(r)
    r.send(r)
    return r

class RoutineException(Exception):
    """
    Special exception type raised from :py:method::`vlcp.event.runnable.RoutineContainer.withException`.
    e.matcher is set to the matcher and e.event is set to the matched event.
    """
    def __init__(self, matcher, event):
        Exception.__init__(self, matcher, event)
        self.matcher = matcher
        self.event = event

class MultipleException(Exception):
    """
    Special exception type raised from :py:method::`vlcp.event.runnable.RoutineContainer.executeAll`
    """
    def __init__(self, exceptions):
        Exception.__init__(self, '%d exceptions occurs in parallel execution' % (len(exceptions),) \
                           + ': ' + repr(exceptions[0]) + ', ...')
        self.exceptions = exceptions

class RoutineContainer(object):
    """
    A routine container groups several routines together and shares data among them. It is also
    used to pass important information like events, matchers and return values to the routine.
    
    Several attributes are commonly used:
        
    currentroutine
        Always set to the executing routine - which is the wrapped routine object of the routine itself
        
    mainroutine
        Set to the main routine `container.main` (started by `container.start`)
        
    """
    
    _container_cache = {}
    
    def __init__(self, scheduler = None, daemon = False):
        """
        Create the routine container.
        
        :param scheduler: The scheduler. This must be set; if None is used, it must be set with
                          `container.bind(scheduler)` before using.
        
        :param daemon: If daemon = True, the main routine `container.main` is set to be a daemon routine.
                       A daemon routine does not stop the scheduler from quitting; if all non-daemon routines
                       are quit, the scheduler stops.
        """
        self.scheduler = scheduler
        self.daemon = daemon
    def bind(self, scheduler):
        """
        If scheduler is not specified, bind the scheduler
        """
        self.scheduler = scheduler
    def main(self):
        """
        The main routine method, should be rewritten to an async method
        """
        raise NotImplementedError
    def start(self, asyncStart = False):
        """
        Start `container.main` as the main routine.
        
        :param asyncStart: if True, start the routine in background. By default, the routine
                           starts in foreground, which means it is executed to the first
                           `yield` statement before returning. If the started routine raises
                           an exception, the exception is re-raised to the caller of `start`
        """
        r = Routine(self.main(), self.scheduler, asyncStart, self, True, self.daemon)
        self.mainroutine = r
        try:
            next(r)
        except StopIteration:
            pass
        return r
    def subroutine(self, iterator, asyncStart = True, name = None, daemon = False):
        """
        Start extra routines in this container.
        
        :param iterator: A coroutine object i.e the return value of an async method `my_routine()`
        
        :param asyncStart: if False, start the routine in foreground. By default, the routine
                           starts in background, which means it is not executed until the current caller
                           reaches the next `yield` statement or quit.
        
        :param name: if not None, `container.<name>` is set to the routine object. This is useful when
                     you want to terminate the routine from outside.
                     
        :param daemon: if True, this routine is set to be a daemon routine.
                       A daemon routine does not stop the scheduler from quitting; if all non-daemon routines
                       are quit, the scheduler stops. 
        """
        r = Routine(iterator, self.scheduler, asyncStart, self, True, daemon)
        if name is not None:
            setattr(self, name, r)
        # Call subroutine may change the currentroutine, we should restore it
        currentroutine = getattr(self, 'currentroutine', None)
        try:
            next(r)
        except StopIteration:
            pass
        self.currentroutine = currentroutine
        return r
    def terminate(self, routine = None):
        """
        Stop a routine.
        
        :param routine: if None, stop the main routine. If not None, it should be a routine object. You
                        can specify a name for a subroutine, and use `container.<name>` to retrieve it.
        """
        if routine is None:
            routine = self.mainroutine
        routine.close()
    def close(self):
        """
        Same as `terminate()`
        """
        self.terminate()
    
    async def wait_for_send(self, event, *, until=None):
        '''
        Send an event to the main event queue. Can call without delegate.
        
        :param until: if the callback returns True, stop sending and return
        
        :return: the last True value the callback returns, or None
        '''
        while True:
            if until:
                r = until()
                if r:
                    return r
            waiter = self.scheduler.send(event)
            if waiter is None:
                break
            await waiter
    waitForSend = wait_for_send
    
    async def wait_with_timeout(self, timeout, *matchers):
        """
        Wait for multiple event matchers, or until timeout.
        
        :param timeout: a timeout value
        
        :param \*matchers: event matchers
        
        :return: (is_timeout, event, matcher). When is_timeout = True, event = matcher = None.
        """
        if timeout is None:
            ev, m = await M_(*matchers)
            return False, ev, m
        else:
            th = self.scheduler.setTimer(timeout)
            try:
                tm = TimerEvent.createMatcher(th)
                ev, m = await M_(*(tuple(matchers) + (tm,)))
                if m is tm:
                    return True, None, None
                else:
                    return False, ev, m
            finally:
                self.scheduler.cancelTimer(th)
    
    waitWithTimeout = wait_with_timeout
    
    async def execute_with_timeout(self, timeout, subprocess):
        """
        Execute a subprocess with timeout. If time limit exceeds, the subprocess is terminated,
        and `is_timeout` is set to True; otherwise the `is_timeout` is set to False.
        
        You can uses `execute_with_timeout` with other help functions to create time limit for them::
        
            timeout, result = await container.execute_with_timeout(10, container.execute_all([routine1(), routine2()]))
        
        :return: (is_timeout, result) When is_timeout = True, result = None
        """
        if timeout is None:
            return (False, await subprocess)
        else:
            th = self.scheduler.setTimer(timeout)
            try:
                tm = TimerEvent.createMatcher(th)
                try:
                    r = await self.with_exception(subprocess, tm)
                except RoutineException as exc:
                    if exc.matcher is tm:
                        return True, None
                    else:
                        raise
                else:
                    return False, r
            finally:
                self.scheduler.cancelTimer(th)
    
    executeWithTimeout = execute_with_timeout
    
    async def do_events(self):
        '''
        Suspend this routine until the next polling. This can be used to give CPU time for other routines and
        socket processings in long calculating procedures. Can call without delegate.
        '''
        self.scheduler.wantContinue()
        await SystemControlEvent.createMatcher(SystemControlEvent.CONTINUE)
    
    doEvents = do_events
    
    async def with_exception(self, subprocess, *matchers):
        """
        Monitoring event matchers while executing a subprocess. If events are matched before the subprocess ends,
        the subprocess is terminated and a RoutineException is raised.
        """
        def _callback(event, matcher):
            raise RoutineException(matcher, event)
        return await self.with_callback(subprocess, _callback, *matchers)
    
    withException = with_exception 
    
    @asyncio.coroutine
    def with_callback(self, subprocess, callback, *matchers, intercept_callback = None):
        """
        Monitoring event matchers while executing a subprocess. `callback(event, matcher)` is called each time
        an event is matched by any event matchers. If the callback raises an exception, the subprocess is terminated.
        
        :param intercept_callback: a callback called before a event is delegated to the inner subprocess
        """
        it_ = _await(subprocess)
        if not matchers and not intercept_callback:
            return (yield from it_)
        try:
            try:
                m = next(it_)
            except StopIteration as e:
                return e.value
            while True:
                if m is None:
                    try:
                        yield
                    except GeneratorExit_:
                        raise
                    except:
                        t,v,tr = sys.exc_info()  # @UnusedVariable
                        try:
                            m = it_.throw(t,v)
                        except StopIteration as e:
                            return e.value
                    else:                        
                        try:
                            m = next(it_)
                        except StopIteration as e:
                            return e.value
                else:
                    while True:
                        try:
                            ev, matcher = yield m + tuple(matchers)
                        except GeneratorExit_:
                            # subprocess is closed in `finally` clause
                            raise
                        except:
                            # delegate this exception inside
                            t,v,tr = sys.exc_info()  # @UnusedVariable
                            try:
                                m = it_.throw(t,v)
                            except StopIteration as e:
                                return e.value
                        else:
                            if matcher in matchers:
                                callback(ev, matcher)
                            else:
                                if intercept_callback:
                                    intercept_callback(ev, matcher)
                                break
                    try:
                        m = it_.send((ev, matcher))
                    except StopIteration as e:
                        return e.value
        finally:
            _close_generator(subprocess)
    
    withCallback = with_callback
    
    async def wait_for_empty(self, queue):
        '''
        Wait for a queue to be empty. Can call without delegate
        '''
        while True:
            m = queue.waitForEmpty()
            if m is None:
                break
            else:
                await m
    
    waitForEmpty = wait_for_empty 
    
    async def wait_for_all(self, *matchers, eventlist = None, eventdict = None, callback = None):
        """
        Wait until each matcher matches an event. When this coroutine method returns,
        `eventlist` is set to the list of events in the arriving order (may not
        be the same as the matchers); `eventdict` is set to a dictionary
        `{matcher1: event1, matcher2: event2, ...}`
        
        :param eventlist: use external event list, so when an exception occurs
                          (e.g. routine close), you can retrieve the result
                          from the passed-in list
        
        :param eventdict: use external event dict
        
        :param callback: if not None, the callback should be a callable callback(event, matcher)
                         which is called each time an event is received
        
        :return: (eventlist, eventdict)
        """
        if eventdict is None:
            eventdict = {}
        if eventlist is None:
            eventlist = []
        ms = len(matchers)
        last_matchers = Diff_(matchers)
        while ms:
            ev, m = await last_matchers
            ms -= 1
            if callback:
                callback(ev, m)
            eventlist.append(ev)
            eventdict[m] = ev
            last_matchers = Diff_(last_matchers, remove=(m,))
        return eventlist, eventdict
    
    waitForAll = wait_for_all
    
    async def wait_for_all_to_process(self, *matchers, eventlist = None, eventdict = None,
                                            callback = None):
        """
        Similar to `waitForAll`, but set `canignore=True` for these events. This ensures
        blocking events are processed correctly.
        """
        def _callback(event, matcher):
            event.canignore = True
            if callback:
                callback(event, matcher)
        return await self.wait_for_all(*matchers, eventlist=eventlist,
                                       eventdict=eventdict, callback=_callback)

    waitForAllToProcess = wait_for_all_to_process
    
    async def wait_for_all_empty(self, *queues):
        """
        Wait for multiple queues to be empty at the same time.

        Require delegate when calling from coroutines running in other containers
        """
        matchers = [m for m in (q.waitForEmpty() for q in queues) if m is not None]
        while matchers:
            await self.wait_for_all(*matchers)
            matchers = [m for m in (q.waitForEmpty() for q in queues) if m is not None]
    
    waitForAllEmpty = wait_for_all_empty
    
    @asyncio.coroutine
    def syscall_noreturn(self, func):
        '''
        Call a syscall method. A syscall method is executed outside of any routines, directly
        in the scheduler loop, which gives it chances to directly operate the event loop.
        See :py:method::`vlcp.event.core.Scheduler.syscall`.
        '''
        matcher = self.scheduler.syscall(func)
        while not matcher:
            yield
            matcher = self.scheduler.syscall(func)
        ev, _ = yield (matcher,)
        return ev
    
    async def syscall(self, func, ignoreException = False):
        """
        Call a syscall method and retrieve its return value
        """
        ev = await self.syscall_noreturn(func)
        if hasattr(ev, 'exception'):
            if ignoreException:
                return
            else:
                raise ev.exception[1]
        else:
            return ev.retvalue
        
    async def delegate(self, subprocess, forceclose = False):
        '''
        Run a subprocess without container support
        
        Many subprocess assume itself running in a specified container, it uses container reference
        like self.events. Calling the subprocess in other containers will fail.
        
        With delegate, you can call a subprocess in any container (or without a container)::
        
            r = await c.delegate(c.someprocess())
        
        :return: original return value
        '''
        finish, r = self.begin_delegate(subprocess)
        return await self.end_delegate(finish, r, forceclose)
    
    async def end_delegate(self, delegate_matcher, routine = None, forceclose = False):
        """
        Retrieve a begin_delegate result. Must be called immediately after begin_delegate
        before any other `await`, or the result might be lost.
        
        Do not use this method without thinking. Always use `RoutineFuture` when possible.
        """
        try:
            ev = await delegate_matcher
            if hasattr(ev, 'exception'):
                raise ev.exception
            else:
                return ev.result
        finally:
            if forceclose and routine:
                routine.close()

    def begin_delegate(self, subprocess):
        '''
        Start the delegate routine, but do not wait for result, instead returns a (matcher, routine) tuple.
        Useful for advanced delegates (e.g. delegate multiple subprocesses in the same time).
        This is NOT a coroutine method.
        
        WARNING: this is not a safe way for asynchronous executing and get the result. Use `RoutineFuture` instead.
        
        :param subprocess: a coroutine
        
        :returns: (matcher, routine) where matcher is a event matcher to get the delegate result, routine is the created routine
        '''
        async def delegateroutine():
            try:
                r = await subprocess
            except:
                _, val, _ = sys.exc_info()
                e = RoutineControlEvent(RoutineControlEvent.DELEGATE_FINISHED, self.currentroutine,
                                        exception=val)
                self.scheduler.emergesend(e)
                raise
            else:
                e = RoutineControlEvent(RoutineControlEvent.DELEGATE_FINISHED, self.currentroutine,
                                        result = r)
                await self.wait_for_send(e)
        r = self.subroutine(generatorwrapper(delegateroutine(), 'subprocess', 'delegate'), True)
        finish = RoutineControlEvent.createMatcher(RoutineControlEvent.DELEGATE_FINISHED, r)
        return finish, r
    
    def begin_delegate_other(self, subprocess, container, retnames = ('',)):
        '''
        DEPRECATED Start the delegate routine, but do not wait for result, instead returns a (matcher routine) tuple.
        Useful for advanced delegates (e.g. delegate multiple subprocesses in the same time).
        This is NOT a coroutine method.
        
        :param subprocess: a coroutine
        
        :param container: container in which to start the routine
        
        :param retnames: get return values from keys. '' for the return value (for compatibility with earlier versions)
        
        :returns: (matcher, routine) where matcher is a event matcher to get the delegate result, routine is the created routine
        '''
        async def delegateroutine():
            try:
                r = await subprocess
            except:
                _, val, _ = sys.exc_info()
                e = RoutineControlEvent(RoutineControlEvent.DELEGATE_FINISHED, container.currentroutine, exception = val)
                container.scheduler.emergesend(e)
                raise
            else:
                e = RoutineControlEvent(RoutineControlEvent.DELEGATE_FINISHED, container.currentroutine,
                                        result = tuple(r if n == '' else getattr(container, n, None)
                                                       for n in retnames))
                await container.waitForSend(e)
        r = container.subroutine(generatorwrapper(delegateroutine(), 'subprocess', 'delegate'), True)
        return (RoutineControlEvent.createMatcher(RoutineControlEvent.DELEGATE_FINISHED, r), r)
    
    beginDelegateOther = begin_delegate_other
    
    async def delegate_other(self, subprocess, container, retnames = ('',), forceclose = False):
        '''
        DEPRECATED Another format of delegate allows delegate a subprocess in another container, and get some returning values
        the subprocess is actually running in 'container'. ::
        
            ret = await self.delegate_other(c.method(), c)
        
        :return: a tuple for retnames values
        
        '''
        finish, r = self.beginDelegateOther(subprocess, container, retnames)
        return await self.end_delegate(finish, r, forceclose)

    delegateOther = delegate_other
    
    async def execute_all_with_names(self, subprocesses, container = None, retnames = ('',), forceclose = True):
        '''
        DEPRECATED Execute all subprocesses and get the return values.
        
        :param subprocesses: sequence of subroutines (coroutines)
        
        :param container: if specified, run subprocesses in another container.
        
        :param retnames: DEPRECATED get return value from container.(name) for each name in retnames.
                         '' for return value (to be compatible with earlier versions)
        
        :param forceclose: force close the routines on exit, so all the subprocesses are terminated
                           on timeout if used with executeWithTimeout
        
        :returns: a list of tuples, one for each subprocess, with value of retnames inside:
                  `[('retvalue1',),('retvalue2',),...]`
        '''
        if not subprocesses:
            return []
        subprocesses = list(subprocesses)
        if len(subprocesses) == 1 and (container is None or container is self) and forceclose:
            # Directly run the process to improve performance
            return [await subprocesses[0]]
        if container is None:
            container = self
        delegates = [self.begin_delegate_other(p, container, retnames) for p in subprocesses]
        matchers = [d[0] for d in delegates]
        try:
            _, eventdict = await self.wait_for_all(*matchers)
            events = [eventdict[m] for m in matchers]
            exceptions = [e.exception for e in events if hasattr(e, 'exception')]
            if exceptions:
                if len(exceptions) == 1:
                    raise exceptions[0]
                else:
                    raise MultipleException(exceptions)
            return [e.result for e in events]
        finally:
            if forceclose:
                for d in delegates:
                    try:
                        container.terminate(d[1])
                    except Exception:
                        pass
    
    executeAll = execute_all_with_names
    
    async def execute_all(self, subprocesses, forceclose=True):
        '''
        Execute all subprocesses and get the return values.
        
        :param subprocesses: sequence of subroutines (coroutines)
        
        :param forceclose: force close the routines on exit, so all the subprocesses are terminated
                           on timeout if used with executeWithTimeout
        
        :returns: a list of return values for each subprocess
        '''
        if not subprocesses:
            return []
        subprocesses = list(subprocesses)
        if len(subprocesses) == 1 and forceclose:
            return [await subprocesses[0]]
        delegates = [self.begin_delegate(p) for p in subprocesses]
        matchers = [d[0] for d in delegates]
        try:
            _, eventdict = await self.wait_for_all(*matchers)
            events = [eventdict[m] for m in matchers]
            exceptions = [e.exception for e in events if hasattr(e, 'exception')]
            if exceptions:
                if len(exceptions) == 1:
                    raise exceptions[0]
                else:
                    raise MultipleException(exceptions)
            return [e.result for e in events]
        finally:
            if forceclose:
                for d in delegates:
                    try:
                        d[1].close()
                    except Exception:
                        pass
    
    @classmethod
    def get_container(cls, scheduler):
        """
        Create temporary instance for helper functions
        """
        if scheduler in cls._container_cache:
            return cls._container_cache[scheduler]
        else:
            c = cls(scheduler)
            cls._container_cache[scheduler] = c
            return c
    
    @classmethod
    def destroy_container_cache(cls, scheduler):
        """
        Remove cached container
        """
        if scheduler in cls._container_cache:
            del cls._container_cache[scheduler]
